#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from pyomo.environ import (
    ConcreteModel,
)

from idaes.core import FlowsheetBlock
from watertap.core.solvers import get_solver

import watertap.property_models.seawater_prop_pack as props
import watertap.property_models.multicomp_aq_sol_prop_pack as props2
from watertap.unit_models.pressure_changer import (
    Pump,
    EnergyRecoveryDevice,
    VariableEfficiency,
)

import idaes.core.util.scaling as iscale
from watertap.unit_models.tests.unit_test_harness import UnitTestHarness

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------


def build_pump():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.SeawaterParameterBlock()

    m.fs.unit = Pump(property_package=m.fs.properties)

    # fully specify system
    feed_flow_mass = 1
    feed_mass_frac_TDS = 0.035
    feed_pressure_in = 1e5
    feed_pressure_out = 5e5
    feed_temperature = 273.15 + 25
    efi_pump = 0.75

    feed_mass_frac_H2O = 1 - feed_mass_frac_TDS
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(
        feed_flow_mass * feed_mass_frac_TDS
    )
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )
    m.fs.unit.inlet.pressure[0].fix(feed_pressure_in)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.outlet.pressure[0].fix(feed_pressure_out)
    m.fs.unit.efficiency_pump.fix(efi_pump)

    iscale.set_scaling_factor(m.fs.unit.control_volume.work[0], 1e-2)
    iscale.set_scaling_factor(m.fs.unit.work_fluid[0], 1e-2)
    iscale.calculate_scaling_factors(m.fs.unit)

    return m


def build_energy_recovery_device():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.SeawaterParameterBlock()

    m.fs.unit = EnergyRecoveryDevice(property_package=m.fs.properties)

    # fully specify system
    feed_flow_mass = 1
    feed_mass_frac_TDS = 0.035
    feed_pressure_in = 5e5
    feed_pressure_out = 1e5
    feed_temperature = 273.15 + 25
    efi_pump = 0.75

    feed_mass_frac_H2O = 1 - feed_mass_frac_TDS
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(
        feed_flow_mass * feed_mass_frac_TDS
    )
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )
    m.fs.unit.inlet.pressure[0].fix(feed_pressure_in)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.outlet.pressure[0].fix(feed_pressure_out)
    m.fs.unit.efficiency_pump.fix(efi_pump)

    iscale.set_scaling_factor(m.fs.unit.control_volume.work[0], 1e-2)
    iscale.set_scaling_factor(m.fs.unit.work_fluid[0], 1e-2)
    iscale.calculate_scaling_factors(m.fs.unit)

    return m


def build_pump_variable_flow():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.SeawaterParameterBlock()

    m.fs.unit = Pump(
        property_package=m.fs.properties,
        variable_efficiency=VariableEfficiency.flow,
    )
    # fully specify system
    feed_flow_mass = 1
    feed_mass_frac_TDS = 0.035
    feed_pressure_in = 1e5
    feed_pressure_out = 5e5
    feed_temperature = 273.15 + 25
    efi_pump = 0.75

    feed_mass_frac_H2O = 1 - feed_mass_frac_TDS
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(
        feed_flow_mass * feed_mass_frac_TDS
    )
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )
    m.fs.unit.inlet.pressure[0].fix(feed_pressure_in)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.outlet.pressure[0].fix(feed_pressure_out)

    m.fs.unit.bep_eta.fix(efi_pump)
    m.fs.unit.flow_ratio.fix(1)

    iscale.set_scaling_factor(m.fs.unit.control_volume.work[0], 1e-2)
    iscale.set_scaling_factor(m.fs.unit.work_fluid[0], 1e-2)
    iscale.set_scaling_factor(m.fs.unit.bep_flow, 1e4)
    iscale.calculate_scaling_factors(m.fs.unit)

    return m


def build_pump_isothermal_energybalancetype_none():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props2.MCASParameterBlock(
        solute_list=["TDS"],
        mw_data={"TDS": 58e-3},
        ignore_neutral_charge=True,
    )

    m.fs.unit = Pump(property_package=m.fs.properties)

    # fully specify system
    feed_flow_mol = 1
    feed_mol_frac_TDS = 0.035
    feed_pressure_in = 1e5
    feed_pressure_out = 5e5
    feed_temperature = 273.15 + 25
    efi_pump = 0.75

    feed_mass_frac_H2O = 1 - feed_mol_frac_TDS
    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "TDS"].fix(
        feed_flow_mol * feed_mol_frac_TDS
    )
    m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mol * feed_mass_frac_H2O
    )
    m.fs.unit.inlet.pressure[0].fix(feed_pressure_in)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.outlet.pressure[0].fix(feed_pressure_out)
    m.fs.unit.efficiency_pump.fix(efi_pump)

    iscale.set_scaling_factor(m.fs.unit.control_volume.work[0], 1e-2)
    iscale.set_scaling_factor(m.fs.unit.work_fluid[0], 1e-2)
    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestPump(UnitTestHarness):
    def configure(self):
        m = build_pump()

        self.unit_solutions[m.fs.unit.work_mechanical[0]] = 521.05643
        self.unit_solutions[m.fs.unit.deltaP[0]] = 4e5
        self.unit_solutions[m.fs.unit.ratioP[0]] = 5
        self.unit_solutions[m.fs.unit.work_fluid[0]] = 390.7923225
        self.unit_solutions[m.fs.unit.control_volume.properties_out[0].temperature] = (
            298.15
        )
        self.unit_solutions[m.fs.unit.control_volume.properties_out[0].enth_flow] = (
            9.931834173e4
        )
        self.unit_solutions[
            m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                "Liq", "H2O"
            ]
        ] = 0.965
        self.unit_solutions[
            m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                "Liq", "TDS"
            ]
        ] = 0.035

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.control_volume.properties_in[0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.control_volume.properties_in[0].flow_mass_phase_comp[
                    "Liq", "TDS"
                ],
                "out": m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                    "Liq", "TDS"
                ],
            },
        }

        return m


class TestEnergyRecoveryDevice(UnitTestHarness):
    def configure(self):
        m = build_energy_recovery_device()

        self.unit_solutions[m.fs.unit.work_mechanical[0]] = -293.094242
        self.unit_solutions[m.fs.unit.deltaP[0]] = -4e5
        self.unit_solutions[m.fs.unit.ratioP[0]] = 0.2
        self.unit_solutions[m.fs.unit.work_fluid[0]] = -390.7923225
        self.unit_solutions[m.fs.unit.control_volume.properties_out[0].temperature] = (
            298.15
        )
        self.unit_solutions[m.fs.unit.control_volume.properties_out[0].enth_flow] = (
            9.89617297e4
        )
        self.unit_solutions[
            m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                "Liq", "H2O"
            ]
        ] = 0.965
        self.unit_solutions[
            m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                "Liq", "TDS"
            ]
        ] = 0.035

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.control_volume.properties_in[0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.control_volume.properties_in[0].flow_mass_phase_comp[
                    "Liq", "TDS"
                ],
                "out": m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                    "Liq", "TDS"
                ],
            },
        }

        return m


class TestPumpVariableFlow(UnitTestHarness):
    def configure(self):
        m = build_pump_variable_flow()

        self.unit_solutions[m.fs.unit.work_mechanical[0]] = 521.05643
        self.unit_solutions[m.fs.unit.deltaP[0]] = 4e5
        self.unit_solutions[m.fs.unit.ratioP[0]] = 5
        self.unit_solutions[m.fs.unit.work_fluid[0]] = 390.7923225
        self.unit_solutions[m.fs.unit.control_volume.properties_out[0].temperature] = (
            298.15
        )
        self.unit_solutions[m.fs.unit.control_volume.properties_out[0].enth_flow] = (
            9.93183417e4
        )
        self.unit_solutions[
            m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                "Liq", "H2O"
            ]
        ] = 0.965
        self.unit_solutions[
            m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                "Liq", "TDS"
            ]
        ] = 0.035

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.control_volume.properties_in[0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.control_volume.properties_in[0].flow_mass_phase_comp[
                    "Liq", "TDS"
                ],
                "out": m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                    "Liq", "TDS"
                ],
            },
        }

        return m


class TestPumpIsothermal(UnitTestHarness):
    def configure(self):
        m = build_pump_isothermal_energybalancetype_none()

        self.unit_solutions[m.fs.unit.work_mechanical[0]] = 10.3466667
        self.unit_solutions[m.fs.unit.deltaP[0]] = 4e5
        self.unit_solutions[m.fs.unit.ratioP[0]] = 5
        self.unit_solutions[m.fs.unit.work_fluid[0]] = 7.76
        self.unit_solutions[m.fs.unit.control_volume.properties_out[0].temperature] = (
            298.15
        )
        self.unit_solutions[
            m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                "Liq", "H2O"
            ]
        ] = 0.01737
        self.unit_solutions[
            m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                "Liq", "TDS"
            ]
        ] = 0.00203

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.control_volume.properties_in[0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.control_volume.properties_in[0].flow_mass_phase_comp[
                    "Liq", "TDS"
                ],
                "out": m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.control_volume.properties_out[0].flow_mass_phase_comp[
                    "Liq", "TDS"
                ],
            },
        }

        return m



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import Constraint
from pyomo.common.config import Bool, ConfigDict, ConfigValue, ConfigBlock, In
from idaes.core import FlowDirection
from idaes.core.util import scaling as iscale

from .MD_channel_base import (
    ConcentrationPolarizationType,
    TemperaturePolarizationType,
    MassTransferCoefficient,
    PressureChangeType,
    FrictionFactor,
)
from idaes.core import (
    FlowDirection,
    EnergyBalanceType,
    MaterialBalanceType,
    MomentumBalanceType,
    useDefault,
    DistributedVars,
)

from idaes.core import declare_process_block_class
from .MD_channel_base import (
    TemperaturePolarizationType,
    MassTransferCoefficient,
    PressureChangeType,
)
from .MD_channel_1D import MDChannel1DBlock
from .membrane_distillation_base import (
    MembraneDistillationBaseData,
    MDconfigurationType,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.logger as idaeslog

__author__ = "Elmira Shamlou"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("MembraneDistillation1D")
class MembraneDistillationData(MembraneDistillationBaseData):
    """
    Standard DCMD Unit Model Class:
    - one dimensional model
    - steady state only
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not.
    **default** - False. Membrane units do not yet support dynamic
    behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. Membrane units do not have defined volume, thus
    this must be False.""",
        ),
    )
    _CONFIG_Template = ConfigBlock()

    _CONFIG_Template.declare(
        "dynamic",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not.
    **default** - False. Membrane units do not yet support dynamic
    behavior.""",
        ),
    )

    _CONFIG_Template.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. Membrane units do not have defined volume, thus
    this must be False.""",
        ),
    )

    _CONFIG_Template.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    _CONFIG_Template.declare(
        "property_package_args",
        ConfigDict(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigDict with arguments to be passed to a property block(s)
and used when constructing these.
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    _CONFIG_Template.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault** - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )

    _CONFIG_Template.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed.
    **default** - useDefault.
    **Valid values:** {
    **EnergyBalanceType.useDefault** - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )

    _CONFIG_Template.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material,
    **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
    **MomentumBalanceType.momentumTotal** - single momentum balance for material,
    **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )

    _CONFIG_Template.declare(
        "flow_direction",
        ConfigValue(
            default=FlowDirection.forward,
            domain=In(FlowDirection),
            description="Direction of flow",
            doc="""
            Options for the direction of flow:

            **default** - ``FlowDirection.forward``

        .. csv-table::
            :header: "Configuration Options", "Description"

            "``FlowDirection.forward``", "Flow is in the forward direction"
            "``FlowDirection.backward``", "Flow is in the backward direction"
            """,
        ),
    )

    _CONFIG_Template.declare(
        "temperature_polarization_type",
        ConfigValue(
            default=TemperaturePolarizationType.calculated,
            domain=In(TemperaturePolarizationType),
            description="External temperature polarization effect",
            doc="""
            Options to account for temperature polarization.

            **default** - ``TemperaturePolarizationType.calculated``

        .. csv-table::
            :header: "Configuration Options", "Description"

            "``TemperaturePolarizationType.none``", "Simplifying assumption to ignore temperature polarization"
            "``TemperaturePolarizationType.fixed``", "Specify an estimated value for the temperature polarization modulus"
            "``TemperaturePolarizationType.calculated``", "Allow model to perform calculation of membrane-interface temperature"
        """,
        ),
    )

    _CONFIG_Template.declare(
        "concentration_polarization_type",
        ConfigValue(
            default=ConcentrationPolarizationType.calculated,
            domain=In(ConcentrationPolarizationType),
            description="External concentration polarization effect in RO",
            doc="""
            Options to account for concentration polarization.

            **default** - ``ConcentrationPolarizationType.calculated``

        .. csv-table::
            :header: "Configuration Options", "Description"

            "``ConcentrationPolarizationType.none``", "Simplifying assumption to ignore concentration polarization"
            "``ConcentrationPolarizationType.fixed``", "Specify an estimated value for the concentration polarization modulus"
            "``ConcentrationPolarizationType.calculated``", "Allow model to perform calculation of membrane-interface concentration"
        """,
        ),
    )

    _CONFIG_Template.declare(
        "mass_transfer_coefficient",
        ConfigValue(
            default=MassTransferCoefficient.calculated,
            domain=In(MassTransferCoefficient),
            description="Mass transfer coefficient in RO feed channel",
            doc="""
            Options to account for mass transfer coefficient.

            **default** - ``MassTransferCoefficient.calculated``

        .. csv-table::
            :header: "Configuration Options", "Description"

            "``MassTransferCoefficient.none``", "Mass transfer coefficient not used in calculations"
            "``MassTransferCoefficient.fixed``", "Specify an estimated value for the mass transfer coefficient in the feed channel"
            "``MassTransferCoefficient.calculated``", "Allow model to perform calculation of mass transfer coefficient"
        """,
        ),
    )

    _CONFIG_Template.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
    constructed,
    **default** - False.
    **Valid values:** {
    **True** - include pressure change terms,
    **False** - exclude pressure change terms.}""",
        ),
    )

    _CONFIG_Template.declare(
        "pressure_change_type",
        ConfigValue(
            default=PressureChangeType.fixed_per_stage,
            domain=In(PressureChangeType),
            description="Pressure change term construction flag",
            doc="""
        Indicates what type of pressure change calculation will be made. To use any of the
        ``pressure_change_type`` options to account for pressure drop, the configuration keyword
        ``has_pressure_change`` must also be set to ``True``. Also, if a value is specified for pressure
        change, it should be negative to represent pressure drop.

        **default** - ``PressureChangeType.fixed_per_stage`` 

        .. csv-table::
            :header: "Configuration Options", "Description"
        
            "``PressureChangeType.fixed_per_stage``", "Specify an estimated value for pressure drop across the membrane feed channel"
            "``PressureChangeType.fixed_per_unit_length``", "Specify an estimated value for pressure drop per unit length across the membrane feed channel"
            "``PressureChangeType.calculated``", "Allow model to perform calculation of pressure drop across the membrane feed channel"
        """,
        ),
    )

    _CONFIG_Template.declare(
        "friction_factor",
        ConfigValue(
            default=FrictionFactor.flat_sheet,
            domain=In(FrictionFactor),
            description="Darcy friction factor correlation",
            doc="""
            Options to account for friction factor correlations.

            **default** - ``FrictionFactor.flat_sheet`` 

        .. csv-table::
            :header: "Configuration Options", "Description"

            "``FrictionFactor.flat_sheet``", "Friction factor correlation for flat-sheet membrane modules"
            "``FrictionFactor.spiral_wound``", "Friction factor correlation for spiral-wound membranes"
        """,
        ),
    )

    _CONFIG_Template.declare(
        "property_package_vapor",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
        **default** - useDefault.
        **Valid values:** {
        **useDefault** - use default package from parent model or flowsheet,
        **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    _CONFIG_Template.declare(
        "property_package_args_vapor",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
        and used when constructing these,
        **default** - None.
        **Valid values:** {
        see property package for documentation.}""",
        ),
    )

    CONFIG.declare("hot_ch", _CONFIG_Template(doc="hot channel config arguments"))
    CONFIG.declare("cold_ch", _CONFIG_Template(doc="cold channel config arguments"))
    CONFIG.declare("gap_ch", _CONFIG_Template(doc="gap channel config arguments"))

    # Common config args for both sides

    CONFIG.declare(
        "MD_configuration_Type",
        ConfigValue(
            default=MDconfigurationType.DCMD,
            domain=In(MDconfigurationType),
            description="Options for selecting membrane distillation process configurations",
            doc="""
            Options for selecting membrane distillation process configurations
            **default** - ``TemperaturePolarizationType.calculated``

        .. csv-table::
            :header: "Configuration Options", "Description"

            "``MDconfigurationType.DCMD``", "Direct Contact Membrane Distillation"
            "``MDconfigurationType.VMD``", "Vacuum Membrane Distillation"
            "``MDconfigurationType.GMD``", "Permeate Gap or Coductive Gap Membrane Distillation"
            "``MDconfigurationType.AGMD``", "Air Gap Membrane Distillation"
        """,
        ),
    )

    CONFIG.declare(
        "area_definition",
        ConfigValue(
            default=DistributedVars.uniform,
            domain=In(DistributedVars),
            description="Argument for defining form of area variable",
            doc="""Argument defining whether area variable should be spatially
        variant or not. **default** - DistributedVars.uniform.
        **Valid values:** {
        DistributedVars.uniform - area does not vary across spatial domain,
        DistributedVars.variant - area can vary over the domain and is indexed
        by time and space.}""",
        ),
    )
    CONFIG.declare(
        "transformation_method",
        ConfigValue(
            default=useDefault,
            description="Discretization method to use for DAE transformation",
            doc="""Discretization method to use for DAE transformation. See
        Pyomo documentation for supported transformations.""",
        ),
    )
    CONFIG.declare(
        "transformation_scheme",
        ConfigValue(
            default=useDefault,
            description="Discretization scheme to use for DAE transformation",
            doc="""Discretization scheme to use when transformating domain. See
        Pyomo documentation for supported schemes.""",
        ),
    )
    CONFIG.declare(
        "finite_elements",
        ConfigValue(
            default=20,
            domain=int,
            description="Number of finite elements length domain",
            doc="""Number of finite elements to use when discretizing length
        domain (default=20)""",
        ),
    )
    CONFIG.declare(
        "collocation_points",
        ConfigValue(
            default=5,
            domain=int,
            description="Number of collocation points per finite element",
            doc="""Number of collocation points to use per finite element when
        discretizing length domain (default=3)""",
        ),
    )

    def _make_MD_channel_control_volume(
        self, name_ch, common_config, individual_config
    ):

        if self.config.transformation_method is useDefault:
            _log.warning(
                "Discretization method was "
                "not specified for the "
                "membrane distillation module. "
                "Defaulting to finite "
                "difference method."
            )
            self.config.transformation_method = "dae.finite_difference"

        if self.config.transformation_scheme is useDefault:
            _log.warning(
                "Discretization scheme was "
                "not specified for the "
                "membrane distillation module."
                "Defaulting to backward finite "
                "difference."
            )
            self.config.transformation_scheme = "BACKWARD"

        if not isinstance(name_ch, str):
            raise TypeError(
                f"{name_ch} is not a string. Please provide a string for the name_ch argument."
            )

        # Build membrane channel control volume
        self.add_component(
            name_ch,
            MDChannel1DBlock(
                dynamic=False,
                has_holdup=False,
                area_definition=common_config.area_definition,
                property_package=individual_config.property_package,
                property_package_args=individual_config.property_package_args,
                transformation_method=common_config.transformation_method,
                transformation_scheme=common_config.transformation_scheme,
                finite_elements=common_config.finite_elements,
                collocation_points=common_config.collocation_points,
            ),
        )

        channel = getattr(self, name_ch)
        if not hasattr(self, "length") and not hasattr(self, "width"):
            self._add_length_and_width()
        channel.add_geometry(
            flow_direction=individual_config.flow_direction,
            length_var=self.length,
            width_var=self.width,
        )

        if not hasattr(self, "eq_area"):
            add_eq_area = True
        else:
            add_eq_area = False
        self._add_area(include_constraint=add_eq_area)

    def _add_mass_transfer(self):
        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            self.config.cold_ch.property_package.phase_list,
            self.config.cold_ch.property_package.component_list,
            doc="Mass transfer from feed to permeate",
        )
        def eq_connect_mass_transfer(b, t, x, p, j):

            if p == "Liq":
                if self.config.MD_configuration_Type == MDconfigurationType.DCMD:
                    return (
                        b.cold_ch.mass_transfer_term[t, x, p, j]
                        == -b.hot_ch.mass_transfer_term[t, x, p, j]
                    )
                else:
                    return Constraint.Skip
            else:
                if self.config.MD_configuration_Type == MDconfigurationType.DCMD:
                    b.cold_ch.mass_transfer_term[t, x, p, j].fix(0)
                    return Constraint.Skip
                elif self.config.MD_configuration_Type == MDconfigurationType.GMD:
                    return Constraint.Skip
                elif self.config.MD_configuration_Type == MDconfigurationType.VMD:
                    b.cold_ch.mass_transfer_term[t, x, "Liq", j].fix(0)
                    return (
                        b.cold_ch.mass_transfer_term[t, x, p, j]
                        == -b.hot_ch.mass_transfer_term[t, x, "Liq", j]
                    )

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            doc="Mass transfer from feed to gap",
        )
        def eq_connect_mass_transfer_gap(b, t, x):

            if self.config.MD_configuration_Type == MDconfigurationType.GMD:
                b.gap_ch.mass_transfer_term[t, x, "Vap", "H2O"].fix(0)
                return (
                    b.gap_ch.mass_transfer_term[t, x, "Liq", "H2O"]
                    == -b.hot_ch.mass_transfer_term[t, x, "Liq", "H2O"]
                )
            else:
                return Constraint.Skip

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            self.config.hot_ch.property_package.phase_list,
            self.config.hot_ch.property_package.component_list,
            doc="Permeate production",
        )
        def eq_permeate_production(b, t, x, p, j):
            if j == "H2O":
                return (
                    b.hot_ch.mass_transfer_term[t, x, p, j]
                    == -b.width * b.flux_mass[t, x]
                )
            else:
                b.hot_ch.mass_transfer_term[t, x, p, j].fix(0)
                return Constraint.Skip

    def _add_heat_transfer(self):
        units_meta = (
            self.config.hot_ch.property_package.get_metadata().get_derived_units
        )

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            doc="Conductive heat transfer to cold channel",
        )
        def eq_conductive_heat_transfer_hot(b, t, x):
            if self.config.MD_configuration_Type == MDconfigurationType.VMD:
                return b.hot_ch.heat[t, x] == -b.width * b.flux_expansion_heat[t, x]
            else:
                return b.hot_ch.heat[t, x] == -b.width * b.flux_conduction_heat[t, x]

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            doc="Conductive heat transfer to cold channel",
        )
        def eq_conductive_heat_transfer_cold(b, t, x):
            if self.config.MD_configuration_Type == MDconfigurationType.DCMD:
                return b.cold_ch.heat[t, x] == -b.hot_ch.heat[t, x]
            elif self.config.MD_configuration_Type == MDconfigurationType.GMD:
                return (
                    b.cold_ch.heat[t, x]
                    == -b.hot_ch.heat[t, x]
                    - b.hot_ch.enthalpy_transfer[t, x]
                    - b.gap_ch.enthalpy_transfer[t, x]
                )
            elif self.config.MD_configuration_Type == MDconfigurationType.VMD:
                return Constraint.Skip

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            doc="Connecting the cold channel conductive heat transfer to conductive heat across the gap",
        )
        def eq_conductive_heat_transfer_gap(b, t, x):
            if self.config.MD_configuration_Type == MDconfigurationType.GMD:
                return (
                    b.cold_ch.heat[t, x] == b.flux_conduction_heat_gap[t, x] * b.width
                )
            else:
                return Constraint.Skip

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            doc="Enthalpy heat transfer from the hot channel",
        )
        def eq_enthalpy_transfer_hot(b, t, x):
            return b.hot_ch.enthalpy_transfer[t, x] == -b.width * b.flux_enth_hot[t, x]

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            doc="Enthalpy heat transfer to the cold channel",
        )
        def eq_enthalpy_transfer_cold(b, t, x):
            if self.config.MD_configuration_Type in [
                MDconfigurationType.DCMD,
            ]:
                return (
                    b.cold_ch.enthalpy_transfer[t, x]
                    == b.width * b.flux_enth_cold[t, x]
                )
            else:
                return Constraint.Skip

        @self.Constraint(
            self.flowsheet().config.time, doc="Gap interface pressure equality"
        )
        def eq_equal_pressure_interface_gap(b, t):
            if self.config.MD_configuration_Type == MDconfigurationType.GMD:
                return (
                    b.gap_ch.properties_interface[
                        t, self.gap_ch.length_domain.first()
                    ].pressure
                    == b.gap_ch.properties[
                        t, self.gap_ch.length_domain.first()
                    ].pressure
                )
            else:
                return Constraint.Skip

        @self.Constraint(
            self.flowsheet().config.time,
            doc="gap bulk temperature in GMD",
        )
        def gap_bulk_temperature_interface_gap(b, t):

            if self.config.MD_configuration_Type == MDconfigurationType.GMD:
                return (
                    b.gap_ch.properties_interface[
                        t, self.gap_ch.length_domain.first()
                    ].temperature
                    == b.gap_ch.properties[
                        t, self.gap_ch.length_domain.first()
                    ].temperature
                )

            else:
                return Constraint.Skip

        @self.Constraint(
            self.flowsheet().config.time,
            doc="gap bulk temperature inlet GMD",
        )
        def gap_bulk_temperature_interface_inlet(b, t):

            if self.config.MD_configuration_Type == MDconfigurationType.GMD:
                if (
                    self.config.gap_ch.temperature_polarization_type
                    == TemperaturePolarizationType.fixed
                ):
                    return (
                        b.hot_ch.properties[
                            t, self.hot_ch.length_domain.first()
                        ].temperature
                        + b.cold_ch.properties[
                            t, self.cold_ch.length_domain.last()
                        ].temperature
                        == b.gap_ch.properties[
                            t, self.gap_ch.length_domain.first()
                        ].temperature
                        * 2
                    )

            else:
                return Constraint.Skip

        if self.config.MD_configuration_Type == MDconfigurationType.GMD:
            for t in self.flowsheet().config.time:
                self.gap_ch.properties_interface[
                    t, self.gap_ch.length_domain.first()
                ].flow_mass_phase_comp["Liq", "H2O"].fix(0)

        @self.Constraint(self.flowsheet().config.time, doc="VMD inlet temperature")
        def eq_vmd_inlet_temp(b, t):
            if self.config.MD_configuration_Type == MDconfigurationType.VMD:
                return (
                    b.cold_ch.properties[
                        t, self.cold_ch.length_domain.first()
                    ].temperature
                    == b.hot_ch.properties[
                        t, self.hot_ch.length_domain.first()
                    ].temperature
                )
            else:
                return Constraint.Skip

    def calculate_scaling_factors(self):

        super().calculate_scaling_factors()

        if self.config.MD_configuration_Type == MDconfigurationType.GMD:
            iscale.set_scaling_factor(
                self.gap_ch.properties_interface[0.0, 0.0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ],
                1e7,
            )
            iscale.set_scaling_factor(
                self.gap_ch.properties_interface[0.0, 0.0].flow_vol_phase["Liq"], 1e10
            )

            for (t, x, p, j), v in self.gap_ch._flow_terms.items():
                iscale.set_scaling_factor(v, 1e3)

            for (t, x), v in self.gap_ch.enthalpy_transfer.items():
                iscale.set_scaling_factor(v, 1e-3)

            for (t, x, p), v in self.gap_ch._enthalpy_flow.items():
                iscale.set_scaling_factor(v, 1e-2)

            for t in self.flowsheet().config.time:
                for x in self.gap_ch.length_domain:
                    for p in self.gap_ch.config.property_package.phase_list:
                        for j in self.gap_ch.config.property_package.component_list:
                            if (p, j) in self.gap_ch.properties[
                                t, x
                            ].flow_mass_phase_comp:
                                v = self.gap_ch.properties[t, x].flow_mass_phase_comp[
                                    p, j
                                ]
                                iscale.set_scaling_factor(v, 1e4)

            for t in self.flowsheet().config.time:
                for x in self.gap_ch.length_domain:
                    for p in self.gap_ch.config.property_package.phase_list:
                        if p in self.gap_ch.properties[t, x].flow_vol_phase:
                            v = self.gap_ch.properties[t, x].flow_vol_phase[p]
                            iscale.set_scaling_factor(v, 1e9)

            for t in self.flowsheet().config.time:
                for x in self.gap_ch.length_domain:
                    for p in self.gap_ch.config.property_package.phase_list:
                        for j in self.gap_ch.config.property_package.component_list:
                            if (p, j) in self.gap_ch.properties_interface[
                                t, x
                            ].flow_mass_phase_comp:
                                var = self.gap_ch.properties_interface[
                                    t, x
                                ].flow_mass_phase_comp[p, j]
                                iscale.set_scaling_factor(var, 1e3)

            for t in self.flowsheet().config.time:
                for x in self.gap_ch.length_domain:
                    for p in self.gap_ch.config.property_package.phase_list:
                        if p in self.gap_ch.properties_interface[t, x].flow_vol_phase:
                            v = self.gap_ch.properties_interface[t, x].flow_vol_phase[p]
                            iscale.set_scaling_factor(v, 1e9)

            for t in self.flowsheet().config.time:
                for x in self.gap_ch.length_domain:
                    for p in self.gap_ch.config.property_package.phase_list:
                        if p in self.gap_ch.properties[t, x].enth_flow_phase:
                            v = self.gap_ch.properties[t, x].enth_flow_phase[p]
                            iscale.set_scaling_factor(v, 1e-3)

        if self.config.MD_configuration_Type == MDconfigurationType.VMD:
            for (t, x, p, j), v in self.cold_ch._flow_terms.items():
                if p == "Vap" and j == "H2O":
                    iscale.set_scaling_factor(v, 1e4)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
ASM1 reaction package.

References:

[1] Henze, M., Grady, C.P.L., Gujer, W., Marais, G.v.R., Matsuo, T.,
"Activated Sludge Model No. 1", 1987, IAWPRC Task Group on Mathematical Modeling
for Design and Operation of Biological Wastewater Treatment
[2] J. Alex, L. Benedetti, J. Copp, K.V. Gernaey, U. Jeppsson, I. Nopens, M.N. Pons,
J.P. Steyer and P. Vanrolleghem, "Benchmark Simulation Model no. 1 (BSM1)", 2018
"""

# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    ReactionParameterBlock,
    ReactionBlockDataBase,
    ReactionBlockBase,
)
from idaes.core.util.misc import add_object_reference
from idaes.core.util.exceptions import BurntToast
import idaes.logger as idaeslog
from idaes.core.scaling import CustomScalerBase, ConstraintScalingScheme
import idaes.core.util.scaling as iscale

# Some more information about this module
__author__ = "Andrew Lee, Xinhong Liu, Adam Atia"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("ASM1ReactionParameterBlock")
class ASM1ReactionParameterData(ReactionParameterBlock):
    """
    Reaction Parameter Block Class
    """

    def build(self):
        """
        Callable method for Block construction.
        """
        super().build()

        self._reaction_block_class = ASM1ReactionBlock

        # Reaction Index
        # Reaction names based on standard numbering in ASM1 paper
        # R1: Aerobic growth of heterotrophs
        # R2: Anoxic growth of heterotrophs
        # R3: Aerobic growth of autotrophs
        # R4: Decay of heterotrophs
        # R5: Decay of autotrophs
        # R6: Ammonification of soluble organic nitrogen
        # R7: Hydrolysis of entrapped organics
        # R8: Hydrolysis of entrapped organic nitrogen
        self.rate_reaction_idx = pyo.Set(
            initialize=["R1", "R2", "R3", "R4", "R5", "R6", "R7", "R8"]
        )

        # Stoichiometric Parameters
        self.Y_A = pyo.Var(
            initialize=0.24,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Yield of cell COD formed per g N consumed, Y_A",
        )
        self.Y_H = pyo.Var(
            initialize=0.67,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Yield of cell COD formed per g COD oxidized, Y_H",
        )

        add_object_reference(self, "f_p", self.config.property_package.f_p)
        add_object_reference(self, "i_xb", self.config.property_package.i_xb)
        add_object_reference(self, "i_xp", self.config.property_package.i_xp)

        # Kinetic Parameters
        self.mu_A = pyo.Var(
            initialize=0.5,
            units=1 / pyo.units.day,
            domain=pyo.PositiveReals,
            doc="Maximum specific growth rate for autotrophic biomass, mu_A",
        )
        self.mu_H = pyo.Var(
            initialize=4,
            units=1 / pyo.units.day,
            domain=pyo.PositiveReals,
            doc="Maximum specific growth rate for heterotrophic biomass, mu_H",
        )
        self.K_S = pyo.Var(
            initialize=10e-3,
            units=pyo.units.kg / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Half-saturation coefficient for heterotrophic biomass, K_S",
        )
        self.K_OH = pyo.Var(
            initialize=0.2e-3,
            units=pyo.units.kg / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Oxygen half-saturation coefficient for heterotrophic biomass, K_O,H",
        )
        self.K_OA = pyo.Var(
            initialize=0.4e-3,
            units=pyo.units.kg / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Oxygen half-saturation coefficient for autotrophic biomass, K_O,A",
        )
        self.K_NO = pyo.Var(
            initialize=0.5e-3,
            units=pyo.units.kg / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Nitrate half-saturation coefficient for denitrifying heterotrophic biomass, K_NO",
        )
        self.b_H = pyo.Var(
            initialize=0.3,
            units=1 / pyo.units.day,
            domain=pyo.PositiveReals,
            doc="Decay coefficient for heterotrophic biomass, b_H",
        )
        self.b_A = pyo.Var(
            initialize=0.05,
            units=1 / pyo.units.day,
            domain=pyo.PositiveReals,
            doc="Decay coefficient for autotrophic biomass, b_A",
        )
        self.eta_g = pyo.Var(
            initialize=0.8,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Correction factor for mu_H under anoxic conditions, eta_g",
        )
        self.eta_h = pyo.Var(
            initialize=0.8,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Correction factor for hydrolysis under anoxic conditions, eta_h",
        )
        self.k_h = pyo.Var(
            initialize=3,
            units=1 / pyo.units.day,
            domain=pyo.PositiveReals,
            doc="Maximum specific hydrolysis rate, k_h",
        )
        self.K_X = pyo.Var(
            initialize=0.1,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Half-saturation coefficient for hydrolysis of slowly biodegradable substrate, K_X",
        )
        self.K_NH = pyo.Var(
            initialize=1e-3,
            units=pyo.units.kg / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Ammonia Half-saturation coefficient for autotrophic biomass, K_NH",
        )
        self.k_a = pyo.Var(
            initialize=0.05e3,
            units=pyo.units.m**3 / pyo.units.kg / pyo.units.day,
            domain=pyo.PositiveReals,
            doc="Ammonification rate, k_a",
        )

        # Reaction Stoichiometry
        # This is the stoichiometric part the Peterson matrix in dict form
        # Note that reaction stoichiometry is on a mass basis.
        # For alkalinity, this requires converting the mass of nitrogen species
        # reacted to mass of alkalinity converted using a charge balance (effectively MW_C/MW_N)
        mw_alk = 12 * pyo.units.kg / pyo.units.kmol
        mw_n = 14 * pyo.units.kg / pyo.units.kmol
        self.rate_reaction_stoichiometry = {
            # R1: Aerobic growth of heterotrophs
            ("R1", "Liq", "H2O"): 0,
            ("R1", "Liq", "S_I"): 0,
            ("R1", "Liq", "S_S"): -1 / self.Y_H,
            ("R1", "Liq", "X_I"): 0,
            ("R1", "Liq", "X_S"): 0,
            ("R1", "Liq", "X_BH"): 1,
            ("R1", "Liq", "X_BA"): 0,
            ("R1", "Liq", "X_P"): 0,
            ("R1", "Liq", "S_O"): -(1 - self.Y_H) / self.Y_H,
            ("R1", "Liq", "S_NO"): 0,
            ("R1", "Liq", "S_NH"): -self.i_xb,
            ("R1", "Liq", "S_ND"): 0,
            ("R1", "Liq", "X_ND"): 0,
            ("R1", "Liq", "S_ALK"): -self.i_xb * (mw_alk / mw_n),
            # R2: Anoxic growth of heterotrophs
            ("R2", "Liq", "H2O"): 0,
            ("R2", "Liq", "S_I"): 0,
            ("R2", "Liq", "S_S"): -1 / self.Y_H,
            ("R2", "Liq", "X_I"): 0,
            ("R2", "Liq", "X_S"): 0,
            ("R2", "Liq", "X_BH"): 1,
            ("R2", "Liq", "X_BA"): 0,
            ("R2", "Liq", "X_P"): 0,
            ("R2", "Liq", "S_O"): 0,
            ("R2", "Liq", "S_NO"): -(1 - self.Y_H) / (2.86 * self.Y_H),
            ("R2", "Liq", "S_NH"): -self.i_xb,
            ("R2", "Liq", "S_ND"): 0,
            ("R2", "Liq", "X_ND"): 0,
            ("R2", "Liq", "S_ALK"): ((1 - self.Y_H) / (2.86 * self.Y_H) - self.i_xb)
            * (mw_alk / mw_n),
            # R3: Aerobic growth of autotrophs
            ("R3", "Liq", "H2O"): 0,
            ("R3", "Liq", "S_I"): 0,
            ("R3", "Liq", "S_S"): 0,
            ("R3", "Liq", "X_I"): 0,
            ("R3", "Liq", "X_S"): 0,
            ("R3", "Liq", "X_BH"): 0,
            ("R3", "Liq", "X_BA"): 1,
            ("R3", "Liq", "X_P"): 0,
            ("R3", "Liq", "S_O"): -(4.57 - self.Y_A) / self.Y_A,
            ("R3", "Liq", "S_NO"): 1 / self.Y_A,
            ("R3", "Liq", "S_NH"): -self.i_xb - 1 / self.Y_A,
            ("R3", "Liq", "S_ND"): 0,
            ("R3", "Liq", "X_ND"): 0,
            ("R3", "Liq", "S_ALK"): (-self.i_xb - 2 / (self.Y_A)) * (mw_alk / mw_n),
            # R4: Decay of heterotrophs
            ("R4", "Liq", "H2O"): 0,
            ("R4", "Liq", "S_I"): 0,
            ("R4", "Liq", "S_S"): 0,
            ("R4", "Liq", "X_I"): 0,
            ("R4", "Liq", "X_S"): 1 - self.f_p,
            ("R4", "Liq", "X_BH"): -1,
            ("R4", "Liq", "X_BA"): 0,
            ("R4", "Liq", "X_P"): self.f_p,
            ("R4", "Liq", "S_O"): 0,
            ("R4", "Liq", "S_NO"): 0,
            ("R4", "Liq", "S_NH"): 0,
            ("R4", "Liq", "S_ND"): 0,
            ("R4", "Liq", "X_ND"): self.i_xb - self.f_p * self.i_xp,
            ("R4", "Liq", "S_ALK"): 0,
            # R5: Decay of autotrophs
            ("R5", "Liq", "H2O"): 0,
            ("R5", "Liq", "S_I"): 0,
            ("R5", "Liq", "S_S"): 0,
            ("R5", "Liq", "X_I"): 0,
            ("R5", "Liq", "X_S"): 1 - self.f_p,
            ("R5", "Liq", "X_BH"): 0,
            ("R5", "Liq", "X_BA"): -1,
            ("R5", "Liq", "X_P"): self.f_p,
            ("R5", "Liq", "S_O"): 0,
            ("R5", "Liq", "S_NO"): 0,
            ("R5", "Liq", "S_NH"): 0,
            ("R5", "Liq", "S_ND"): 0,
            ("R5", "Liq", "X_ND"): self.i_xb - self.f_p * self.i_xp,
            ("R5", "Liq", "S_ALK"): 0,
            # R6: Ammonification of soluble organic nitrogen
            ("R6", "Liq", "H2O"): 0,
            ("R6", "Liq", "S_I"): 0,
            ("R6", "Liq", "S_S"): 0,
            ("R6", "Liq", "X_I"): 0,
            ("R6", "Liq", "X_S"): 0,
            ("R6", "Liq", "X_BH"): 0,
            ("R6", "Liq", "X_BA"): 0,
            ("R6", "Liq", "X_P"): 0,
            ("R6", "Liq", "S_O"): 0,
            ("R6", "Liq", "S_NO"): 0,
            ("R6", "Liq", "S_NH"): 1,
            ("R6", "Liq", "S_ND"): -1,
            ("R6", "Liq", "X_ND"): 0,
            ("R6", "Liq", "S_ALK"): 1 * (mw_alk / mw_n),
            # R7: Hydrolysis of entrapped organics
            ("R7", "Liq", "H2O"): 0,
            ("R7", "Liq", "S_I"): 0,
            ("R7", "Liq", "S_S"): 1,
            ("R7", "Liq", "X_I"): 0,
            ("R7", "Liq", "X_S"): -1,
            ("R7", "Liq", "X_BH"): 0,
            ("R7", "Liq", "X_BA"): 0,
            ("R7", "Liq", "X_P"): 0,
            ("R7", "Liq", "S_O"): 0,
            ("R7", "Liq", "S_NO"): 0,
            ("R7", "Liq", "S_NH"): 0,
            ("R7", "Liq", "S_ND"): 0,
            ("R7", "Liq", "X_ND"): 0,
            ("R7", "Liq", "S_ALK"): 0,
            # R8: Hydrolysis of entrapped organic nitrogen
            ("R8", "Liq", "H2O"): 0,
            ("R8", "Liq", "S_I"): 0,
            ("R8", "Liq", "S_S"): 0,
            ("R8", "Liq", "X_I"): 0,
            ("R8", "Liq", "X_S"): 0,
            ("R8", "Liq", "X_BH"): 0,
            ("R8", "Liq", "X_BA"): 0,
            ("R8", "Liq", "X_P"): 0,
            ("R8", "Liq", "S_O"): 0,
            ("R8", "Liq", "S_NO"): 0,
            ("R8", "Liq", "S_NH"): 0,
            ("R8", "Liq", "S_ND"): 1,
            ("R8", "Liq", "X_ND"): -1,
            ("R8", "Liq", "S_ALK"): 0,
        }
        # Fix all the variables we just created
        for v in self.component_objects(pyo.Var, descend_into=False):
            v.fix()

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties(
            {
                "reaction_rate": {"method": "_rxn_rate"},
            }
        )
        obj.add_default_units(
            {
                "time": pyo.units.s,
                "length": pyo.units.m,
                "mass": pyo.units.kg,
                "amount": pyo.units.kmol,
                "temperature": pyo.units.K,
            }
        )


class ASM1ReactionScaler(CustomScalerBase):
    """
    Scaler for the Activated Sludge Model No.1 reaction package.

    Variables are scaled by their default scaling factor (if no user input provided), and constraints
    are scaled using the inverse maximum scheme.
    """

    # TODO: Revisit this scaling factor
    DEFAULT_SCALING_FACTORS = {"reaction_rate": 1e2}

    def variable_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):

        if model.is_property_constructed("reaction_rate"):
            for j in model.reaction_rate.values():
                self.scale_variable_by_default(j, overwrite=overwrite)

    def constraint_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        # TODO: Revisit this scaling methodology
        # Consider scale_constraint_by_default or scale_constraints_by_jacobian_norm
        if model.is_property_constructed("rate_expression"):
            for j in model.rate_expression.values():
                self.scale_constraint_by_nominal_value(
                    j,
                    scheme=ConstraintScalingScheme.inverseMaximum,
                    overwrite=overwrite,
                )


class _ASM1ReactionBlock(ReactionBlockBase):
    """
    This Class contains methods which should be applied to Reaction Blocks as a
    whole, rather than individual elements of indexed Reaction Blocks.
    """

    default_scaler = ASM1ReactionScaler

    def initialize(self, outlvl=idaeslog.NOTSET, **kwargs):
        """
        Initialization routine for reaction package.

        Keyword Arguments:
            outlvl : sets output level of initialization routine

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        init_log.info("Initialization Complete.")


@declare_process_block_class("ASM1ReactionBlock", block_class=_ASM1ReactionBlock)
class ASM1ReactionBlockData(ReactionBlockDataBase):
    """
    ReactionBlock for ASM1.
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super().build()

        # Create references to state vars
        # Concentration
        add_object_reference(self, "conc_mass_comp_ref", self.state_ref.conc_mass_comp)

    # Rate of reaction method
    def _rxn_rate(self):
        self.reaction_rate = pyo.Var(
            self.params.rate_reaction_idx,
            initialize=0,
            doc="Rate of reaction",
            units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
        )

        try:

            def rate_expression_rule(b, r):
                if r == "R1":
                    # R1: Aerobic growth of heterotrophs
                    return b.reaction_rate[r] == pyo.units.convert(
                        b.params.mu_H
                        * (
                            b.conc_mass_comp_ref["S_S"]
                            / (b.params.K_S + b.conc_mass_comp_ref["S_S"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_O"]
                            / (b.params.K_OH + b.conc_mass_comp_ref["S_O"])
                        )
                        * b.conc_mass_comp_ref["X_BH"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R2":
                    # R2: Anoxic growth of heterotrophs
                    return b.reaction_rate[r] == pyo.units.convert(
                        b.params.mu_H
                        * (
                            b.conc_mass_comp_ref["S_S"]
                            / (b.params.K_S + b.conc_mass_comp_ref["S_S"])
                        )
                        * (
                            b.params.K_OH
                            / (b.params.K_OH + b.conc_mass_comp_ref["S_O"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_NO"]
                            / (b.params.K_NO + b.conc_mass_comp_ref["S_NO"])
                        )
                        * b.params.eta_g
                        * b.conc_mass_comp_ref["X_BH"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R3":
                    # R3: Aerobic growth of autotrophs
                    return b.reaction_rate[r] == pyo.units.convert(
                        b.params.mu_A
                        * (
                            b.conc_mass_comp_ref["S_NH"]
                            / (b.params.K_NH + b.conc_mass_comp_ref["S_NH"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_O"]
                            / (b.params.K_OA + b.conc_mass_comp_ref["S_O"])
                        )
                        * b.conc_mass_comp_ref["X_BA"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R4":
                    # R4: Decay of heterotrophs
                    return b.reaction_rate[r] == pyo.units.convert(
                        b.params.b_H * b.conc_mass_comp_ref["X_BH"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R5":
                    # R5: Decay of autotrophs
                    return b.reaction_rate[r] == pyo.units.convert(
                        b.params.b_A * b.conc_mass_comp_ref["X_BA"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R6":
                    # R6: Ammonification of soluble organic nitrogen
                    return b.reaction_rate[r] == pyo.units.convert(
                        b.params.k_a
                        * b.conc_mass_comp_ref["S_ND"]
                        * b.conc_mass_comp_ref["X_BH"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R7":
                    # R7: Hydrolysis of entrapped organics
                    return b.reaction_rate[r] == pyo.units.convert(
                        b.params.k_h
                        * b.conc_mass_comp_ref["X_S"]
                        / (
                            b.params.K_X * b.conc_mass_comp_ref["X_BH"]
                            + b.conc_mass_comp_ref["X_S"]
                        )
                        * (
                            (
                                b.conc_mass_comp_ref["S_O"]
                                / (b.params.K_OH + b.conc_mass_comp_ref["S_O"])
                            )
                            + b.params.eta_h
                            * b.params.K_OH
                            / (b.params.K_OH + b.conc_mass_comp_ref["S_O"])
                            * (
                                b.conc_mass_comp_ref["S_NO"]
                                / (b.params.K_NO + b.conc_mass_comp_ref["S_NO"])
                            )
                        )
                        * b.conc_mass_comp_ref["X_BH"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R8":
                    # R8: Hydrolysis of entrapped organic nitrogen
                    return b.reaction_rate[r] == (
                        b.reaction_rate["R7"]
                        * (
                            b.conc_mass_comp_ref["X_ND"]
                            / (
                                b.conc_mass_comp_ref["X_S"]
                                + 1e-10 * pyo.units.kg / pyo.units.m**3
                            )
                        )
                    )
                else:
                    raise BurntToast()

            self.rate_expression = pyo.Constraint(
                self.params.rate_reaction_idx,
                rule=rate_expression_rule,
                doc="ASM1 rate expressions",
            )

        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.reaction_rate)
            self.del_component(self.rate_expression)
            raise

    def get_reaction_rate_basis(self):
        return MaterialFlowBasis.mass

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        iscale.constraint_scaling_transform(self.rate_expression["R5"], 1e3)
        iscale.constraint_scaling_transform(self.rate_expression["R3"], 1e3)
        iscale.constraint_scaling_transform(self.rate_expression["R4"], 1e3)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from watertap.costing.util import (
    register_costing_parameter_block,
    make_capital_cost_var,
)


def build_surrogate_crystallizer_cost_param_block(blk):

    blk.steam_pressure = pyo.Var(
        initialize=3,
        units=pyo.units.bar,
        doc="Steam pressure (gauge) for crystallizer heating: 3 bar default based on Dutta example",
    )

    blk.efficiency_pump = pyo.Var(
        initialize=0.7,
        units=pyo.units.dimensionless,
        doc="Crystallizer pump efficiency - assumed",
    )

    blk.pump_head_height = pyo.Var(
        initialize=1,
        units=pyo.units.m,
        doc="Crystallizer pump head height -  assumed, unvalidated",
    )

    # Crystallizer operating cost information from literature
    blk.fob_unit_cost = pyo.Var(
        initialize=675000,
        doc="Forced circulation crystallizer reference free-on-board cost (Woods, 2007)",
        units=pyo.units.USD_2007,
    )

    blk.ref_capacity = pyo.Var(
        initialize=1,
        doc="Forced circulation crystallizer reference crystal capacity (Woods, 2007)",
        units=pyo.units.kg / pyo.units.s,
    )

    blk.ref_exponent = pyo.Var(
        initialize=0.53,
        doc="Forced circulation crystallizer cost exponent factor (Woods, 2007)",
        units=pyo.units.dimensionless,
    )

    blk.iec_percent = pyo.Var(
        initialize=1.43,
        doc="Forced circulation crystallizer installed equipment cost (Diab and Gerogiorgis, 2017)",
        units=pyo.units.dimensionless,
    )

    blk.steam_cost = pyo.Var(
        initialize=0.004,
        units=pyo.units.USD_2018 / (pyo.units.meter**3),
        doc="Steam cost, Panagopoulos (2019)",
    )

    costing = blk.parent_block()
    costing.register_flow_type("steam", blk.steam_cost)


def cost_surrogate_crystallizer(blk):
    """
    Function for costing the surrogate crystallizer by the mass flow of produced crystals.
    The operating cost model assumes that heat is supplied via condensation of saturated steam (see Dutta et al.)
    """
    cost_crystallizer_by_crystal_mass(blk)


def _cost_crystallizer_flows(blk):
    blk.costing_package.cost_flow(
        pyo.units.convert(
            (blk.unit_model.heat_required / _compute_steam_properties(blk)),
            to_units=pyo.units.m**3 / pyo.units.s,
        ),
        "steam",
    )


@register_costing_parameter_block(
    build_rule=build_surrogate_crystallizer_cost_param_block,
    parameter_block_name="surrogate_crystallizer",
)
def cost_crystallizer_by_crystal_mass(blk):
    """
    Mass-based capital cost for FC crystallizer
    """
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.cost_factor = 1  # blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            (
                blk.costing_package.surrogate_crystallizer.iec_percent
                * blk.costing_package.surrogate_crystallizer.fob_unit_cost
                * (
                    blk.unit_model.flow_mass_sol_total
                    / blk.costing_package.surrogate_crystallizer.ref_capacity
                )
                ** blk.costing_package.surrogate_crystallizer.ref_exponent
            ),
            to_units=blk.costing_package.base_currency,
        )
    )
    _cost_crystallizer_flows(blk)


def _compute_steam_properties(blk):
    """
    Function for computing saturated steam properties for thermal heating estimation.

    Args:
        pressure_sat:   Steam gauge pressure in bar

    Out:
        Steam thermal capacity (latent heat of condensation * density) in kJ/m3
    """
    pressure_sat = blk.costing_package.surrogate_crystallizer.steam_pressure
    # 1. Compute saturation temperature of steam: computed from El-Dessouky expression
    tsat_constants = [
        42.6776 * pyo.units.K,
        -3892.7 * pyo.units.K,
        1000 * pyo.units.kPa,
        -9.48654 * pyo.units.dimensionless,
    ]
    psat = (
        pyo.units.convert(pressure_sat, to_units=pyo.units.kPa)
        + 101.325 * pyo.units.kPa
    )
    temperature_sat = tsat_constants[0] + tsat_constants[1] / (
        pyo.log(psat / tsat_constants[2]) + tsat_constants[3]
    )

    # 2. Compute latent heat of condensation/vaporization: computed from Sharqawy expression
    t = temperature_sat - 273.15 * pyo.units.K
    enth_mass_units = pyo.units.J / pyo.units.kg
    t_inv_units = pyo.units.K**-1
    dh_constants = [
        2.501e6 * enth_mass_units,
        -2.369e3 * enth_mass_units * t_inv_units**1,
        2.678e-1 * enth_mass_units * t_inv_units**2,
        -8.103e-3 * enth_mass_units * t_inv_units**3,
        -2.079e-5 * enth_mass_units * t_inv_units**4,
    ]
    dh_vap = (
        dh_constants[0]
        + dh_constants[1] * t
        + dh_constants[2] * t**2
        + dh_constants[3] * t**3
        + dh_constants[4] * t**4
    )
    dh_vap = pyo.units.convert(dh_vap, to_units=pyo.units.kJ / pyo.units.kg)

    # 3. Compute specific volume: computed from Affandi expression (Eq 5)
    t_critical = 647.096 * pyo.units.K
    t_red = temperature_sat / t_critical  # Reduced temperature
    sp_vol_constants = [
        -7.75883 * pyo.units.dimensionless,
        3.23753 * pyo.units.dimensionless,
        2.05755 * pyo.units.dimensionless,
        -0.06052 * pyo.units.dimensionless,
        0.00529 * pyo.units.dimensionless,
    ]
    log_sp_vol = (
        sp_vol_constants[0]
        + sp_vol_constants[1] * (pyo.log(1 / t_red)) ** 0.4
        + sp_vol_constants[2] / (t_red**2)
        + sp_vol_constants[3] / (t_red**4)
        + sp_vol_constants[4] / (t_red**5)
    )
    sp_vol = pyo.exp(log_sp_vol) * pyo.units.m**3 / pyo.units.kg

    # 4. Return specific energy: density * latent heat
    return dh_vap / sp_vol



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

###############################################################################
#
# OLI Systems, Inc. Copyright © 2022, all rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# 3. Neither the name of OLI Systems, Inc. nor the names of any contributors to
# the software made available herein may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# You are under no obligation whatsoever to provide any bug fixes, patches, or upgrades to the
# features, functionality or performance of the source code ("Enhancements") to anyone; however,
# if you choose to make your Enhancements available either publicly, or directly to OLI Systems, Inc.,
# without imposing a separate written license agreement for such Enhancements, then you hereby grant
# the following license: a non-exclusive, royalty-free perpetual license to install, use, modify, prepare
# derivative works, incorporate into other computer software, distribute, and sublicense such enhancements
# or derivative works thereof, in binary and source code form.
###############################################################################
__author__ = "Oluwamayowa Amusat, Alexander Dudchenko, Paul Vecchiarelli, Adam Atia"


import logging

import json
from pathlib import Path

from copy import deepcopy
from itertools import product

from watertap.tools.oli_api.util.watertap_to_oli_helper_functions import (
    get_oli_name,
    get_charge,
    get_charge_group,
)
from watertap.tools.oli_api.util.fixed_keys_dict import (
    optional_properties,
    input_unit_set,
    output_unit_set,
)

from numpy import reshape, sqrt

_logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter(
    "OLIAPI - %(asctime)s - %(levelname)s - %(message)s", "%H:%M:%S"
)
handler.setFormatter(formatter)
_logger.addHandler(handler)
_logger.setLevel(logging.DEBUG)


class Flash:
    """
    A class to execute OLI Cloud flash calculations.

    :param optional_properties: dictionary for optional properties to attach to OLI calls, all True by default
    :param input_unit_set: dictionary for conversions between OLI and Pyomo unit names
    :param output_unit_set: dictionary for preferred output units
    :param relative_inflows: bool switch for surveys - true to add specified value to initial value, false to replace initial value with specified value
    :param debug_level: string defining level of logging activity
    """

    def __init__(
        self,
        optional_properties=optional_properties,
        input_unit_set=input_unit_set,
        output_unit_set=output_unit_set,
        relative_inflows=True,
        debug_level="INFO",
    ):
        self.optional_properties = optional_properties
        self.input_unit_set = input_unit_set
        self.output_unit_set = output_unit_set
        self.relative_inflows = relative_inflows

        if debug_level == "INFO":
            _logger.setLevel(logging.INFO)
        else:
            _logger.setLevel(logging.DEBUG)

    def configure_water_analysis(
        self,
        inflows=None,
        temperature=None,
        pressure=None,
        reconciliation=None,
        electroneutrality=None,
        makeup_ion=None,
        ph=None,
        acid_titrant=None,
        base_titrant=None,
        alkalinity=None,
        alkalinity_ph=None,
        alkalinity_titrant=None,
        tic=None,
        allow_solids=False,
        included_solids=None,
        excluded_solids=None,
        calc_alkalinity=False,
        use_scaling_rigorous=True,
        file_name=None,
    ):
        """
        Configure Water Analysis JSON input.

        :param inflows: dictionary of solutes
        :param temperature: float for temperature in Kelvins
        :param pressure: float for pressure in Pascals
        :param reconciliation: string for method of reconciliation: "EquilCalcOnly" (default), "ReconcilePh", "ReconcilePhAndAlkalinity", or "ReconcilePhAndAlkalinityAndTic"; "ReconcileCo2Gas" not supported currently.
        :param electroneutrality: string for method of electroneutrality calculation: "DominantIon", "ProrateCations", "ProrateAnions", "Prorate", "AutoNACL", or "MakeupIon" are supported
        :param makeup_ion: string for ion to use for electroneutrality balance, if "MakeupIon,
        :param ph: float for pH to reconcile solution to, required for pH based reconciliation
        :param acid_titrant: string for acidification titrant, used in pH based reconciliation
        :param base_titrant: string for basification titrant, used in pH based reconciliation
        :param alkalinity: float for alkalinity to reconcile solution to, required for Alk based reconciliation
        :param alkalinity_ph: float for alkalinity endpoint ph, used in Alk based reconciliation
        :param alkalinity_titrant: string for alkalinity titration species, used in Alk based reconciliation
        :param tic: float for total inorganic carbon concentration to reconcile solution to, required for TIC based reconcilation
        :param allow_solids: bool to enable solid phase formation
        :param included_solids: list of solids to include in analysis
        :param excluded_solids: list of solids to exclude from analysis
        :param calc_alkalinity: bool to calculate alkalinity of solution
        :param use_scaling_rigorous: bool to switch between Rigorous (default) and Estimated scaling computations
        :param file_name: string for file to write, if any
        :param mesh_grid: if True (default) the input array will be combined to generate combination of all possible samples
            if False, the direct values in survey_arrays will be used

        :return json_input: JSON for Water Analysis
        """

        _logger.info("Configuring Water Analysis JSON ...")
        input_list = []

        if not inflows:
            raise RuntimeError("Inflows must be defined for Water Analysis.")

        temp_input = {
            "group": "Properties",
            "name": "Temperature",
            "unit": self.input_unit_set["temperature"]["oli_unit"],
            "value": 273.15,
        }
        if temperature is not None:
            if float(temperature):
                temp_input.update({"value": float(temperature)})
            else:
                raise ValueError(f"Invalid temperature: {temperature}. Expected number")
        input_list.append(temp_input)

        pres_input = {
            "group": "Properties",
            "name": "Pressure",
            "unit": self.input_unit_set["pressure"]["oli_unit"],
            "value": 101325,
        }
        if pressure is not None:
            if float(pressure):
                pres_input.update({"value": float(pressure)})
            else:
                raise ValueError(f"Invalid pressure: {pressure}. Expected number")
        input_list.append(pres_input)

        reconciliation_options = [
            "EquilCalcOnly",
            "ReconcilePh",
            "ReconcilePhAndAlkalinity",
            "ReconcilePhAndAlkalinityAndTic",
        ]
        rec_input = {
            "group": "Calculation Options",
            "name": "CalcType",
            "value": "EquilCalcOnly",
        }
        if reconciliation is not None:
            if reconciliation in reconciliation_options:
                rec_input.update({"value": reconciliation})
            else:
                raise RuntimeError(
                    f"Invalid reconciliation option: {reconciliation}."
                    + f" Use one of {reconciliation_options}"
                )
        else:
            reconciliation = "EquilCalcOnly"

        input_list.append(rec_input)
        additional_req_input = []
        additional_req_args = []
        if "Ph" in reconciliation:
            additional_req_args.append([ph, acid_titrant, base_titrant])
            if not acid_titrant:
                acid_titrant = "HCl"
            if not base_titrant:
                base_titrant = "NaOH"
            additional_req_input.extend(
                [
                    {
                        "group": "Properties",
                        "name": "pH",
                        "value": ph,
                    },
                    {
                        "group": "Calculation Options",
                        "name": "PhAcidTitrant",
                        "value": get_oli_name(acid_titrant),
                    },
                    {
                        "group": "Calculation Options",
                        "name": "PhBaseTitrant",
                        "value": get_oli_name(base_titrant),
                    },
                ]
            )
        if "Alk" in reconciliation:
            additional_req_args.append([alkalinity, alkalinity_ph, alkalinity_titrant])
            if not alkalinity_titrant:
                alkalinity_titrant = "H2SO4"
            if not alkalinity_ph:
                alkalinity_ph = 4.5
                _logger.info("No alkalinity endpoint pH specified. Assuming 4.5.")
                additional_req_input.extend(
                    [
                        {
                            "group": "Properties",
                            "name": "Alkalinity",
                            "unit": self.input_unit_set["alkalinity"]["oli_unit"],
                            "value": alkalinity,
                        },
                        {
                            "group": "Properties",
                            "name": "AlkalinityTitrationEndPointpH",
                            "value": alkalinity_ph,
                        },
                        {
                            "group": "Calculation Options",
                            "name": "AlkalinityPhTitrant",
                            "value": alkalinity_titrant,
                        },
                    ]
                )
        if "Tic" in reconciliation:
            additional_req_args.append([tic])
            additional_req_input.append(
                {
                    "group": "Properties",
                    "name": "TIC",
                    "unit": self.input_unit_set["TIC"]["oli_unit"],
                    "value": tic,
                }
            )
        missing_keys = [arg for arg in additional_req_args if arg is None]
        if missing_keys:
            raise RuntimeError(f"Missing keys for {reconciliation}: {missing_keys}")
        input_list.extend(additional_req_input)

        electroneutrality_options = [
            "DominantIon",
            "ProrateCations",
            "ProrateAnions",
            "Prorate",
            "AutoNACL",
            "MakeupIon",
        ]
        elec_input = {
            "group": "Electroneutrality Options",
            "name": "ElectroNeutralityBalanceType",
            "value": "DominantIon",
        }

        if electroneutrality is not None:
            if electroneutrality in electroneutrality_options:
                elec_input.update({"value": electroneutrality})
            else:
                raise RuntimeError(
                    f"Invalid reconciliation option: {electroneutrality}."
                    + f" Use one of {electroneutrality_options}"
                )
        input_list.append(elec_input)
        if electroneutrality == "MakeupIon":
            if makeup_ion is not None:
                input_list.append(
                    {
                        "group": "Electroneutrality Options",
                        "name": "MakeupIonBaseTag",
                        "value": get_oli_name(makeup_ion),
                    }
                )

        input_list.extend(
            [
                {
                    "group": "Calculation Options",
                    "name": "AllowSolidsToForm",
                    "value": bool(allow_solids),
                },
                {
                    "group": "Calculation Options",
                    "name": "CalcAlkalnity",
                    "value": bool(calc_alkalinity),
                },
            ]
        )
        conc_unit = self.input_unit_set["inflows"]["oli_unit"]
        _logger.info(f"Using {conc_unit} for inflows input")
        for k, v in inflows.items():
            charge = get_charge(k)
            input_list.append(
                {
                    "group": get_charge_group(charge),
                    "name": get_oli_name(k),
                    "unit": conc_unit,
                    "value": v,
                    "charge": charge,
                }
            )

        json_input = self._add_to_json(
            "wateranalysis",
            input_list,
            included_solids,
            excluded_solids,
            use_scaling_rigorous,
            file_name,
        )
        return json_input

    def configure_flash_analysis(
        self,
        inflows=None,
        flash_method=None,
        temperature=None,
        pressure=None,
        calculated_variable=None,
        enthalpy=None,
        vapor_amount=None,
        vapor_fraction=None,
        volume=None,
        ph=None,
        acid_titrant=None,
        base_titrant=None,
        formed_solid=None,
        precipitant_inflow=None,
        included_solids=None,
        excluded_solids=None,
        contact_surface=None,
        flow_type=None,
        diameter=None,
        liq_velocity=None,
        gas_velocity=None,
        rot_velocity=None,
        shear_stress=None,
        roughness=None,
        nonaqueous_visc=None,
        water_cut_inversion=None,
        relative_visc_inversion=None,
        use_scaling_rigorous=True,
        file_name=None,
    ):
        """
        Configure Flash Analysis JSON input.

        :param inflows: dictionary of solutes, of the form {"unit": unit, "values": {solute_name: concentration_value}}; otherwise, {solute_name: concentration_value}
        :param flash_method: string for flash calculation name
        :param temperature: float for temperature in Kelvins
        :param pressure: float for pressure in Pascals
        :param calculated_variable: string for variable to calculate, such as temperature or pressure, used in 'bubblepoint', 'dewpoint', 'vapor-amount', 'vapor-fraction', and 'isochoric' flashes
        :param enthalpy: float for total enthalpy in Joules, used in 'isenthalpic' flash
        :param vapor_amount: float for vapor phase Moles, used in 'vapor-amount' flash
        :param vapor_fraction: float for vapor phase in Mole %, used in 'vapor-fraction' flash
        :param volume: float for total volume in Cubic Meters, used in 'isochoric' flash
        :param ph: float for target pH, used in 'setph' flash
        :param acid_titrant: string for acidification titrant, used in 'setph' flash
        :param base_titrant: string for basification titrant, used in 'setph' flash
        :param formed_solid: string for solid species to precipitate based on inflow sweep, used in 'precipitation-point'
        :param precipitant_inflow: string for inflow species to sweep, used in 'precipitation-point'
        :param included_solids: list of solids to include in analysis
        :param excluded_solids: list of solids to exclude from analysis
        :param contact_surface: string for contact surface metal name
        :param flow_type: string for flow configuration
        :param diameter: float for diameter of surface (i.e., pipe or rotor)
        :param liq_velocity: float for velocity of liquid flow
        :param gas_velocity: float for velocity of vapor flow, used in 'approximateMultiPhaseFlow'
        :param rot_velocity: float for rotational velocity
        :param shear_stress: float for defined shear stress, used in 'definedShearStress'
        :param roughness: float for pipe roughness, used in 'approximateMultiPhaseFlow'
        :param nonaqueous_visc: float for absolute viscosity of nonaqueous phase, used in 'approximateMultiPhaseFlow'
        :param water_cut_inversion: float for water cut at point of dispersion inversion, used in 'approximateMultiPhaseFlow'
        :param relative_visc_inversion: float for maximum relative viscosity of dispersion at inversion, used in 'approximateMultiPhaseFlow'
        :param use_scaling_rigorous: bool to switch between Rigorous (default) and Estimated scaling computations
        :param file_name: string for file to write, if any

        :return json_input: JSON for Water Analysis
        """

        _logger.info(f"Configuring {flash_method} Flash JSON ...")

        if flash_method not in [
            "isothermal",
            "isenthalpic",
            "bubblepoint",
            "dewpoint",
            "vapor-amount",
            "vapor-fraction",
            "isochoric",
            "setph",
            "precipitation-point",
            "corrosion-rates",
        ]:
            raise RuntimeError(
                f"Failed to configure Flash. Invalid method: {flash_method}"
            )

        if not inflows:
            raise RuntimeError("Inflows must be defined for Flash Analysis.")

        input_dict = {}
        temp_input = {
            "unit": self.input_unit_set["temperature"]["oli_unit"],
            "value": 273.15,
        }
        if temperature:
            if float(temperature):
                temp_input.update({"value": float(temperature)})
            else:
                raise ValueError(f"Invalid temperature: {temperature}. Expected number")
        input_dict["temperature"] = temp_input

        pres_input = {
            "unit": self.input_unit_set["pressure"]["oli_unit"],
            "value": 101325,
        }
        if pressure:
            if float(pressure):
                pres_input.update({"value": float(pressure)})
            else:
                raise ValueError(f"Invalid pressure: {pressure}. Expected number")
        input_dict["pressure"] = pres_input

        if flash_method in [
            "bubblepoint",
            "dewpoint",
            "vapor-amount",
            "vapor-fraction",
            "isochoric",
        ]:

            if calculated_variable is not None:
                if calculated_variable not in ["temperature", "pressure"]:
                    raise RuntimeError(
                        f"Invalid input for 'calculated_variable': {calculated_variable}; 'temperature' or 'pressure' supported."
                    )
                _logger.info(
                    f"{flash_method} will calculate {calculated_variable} as its variable"
                )
                input_dict["calculatedVariable"] = calculated_variable
            else:
                raise RuntimeError(
                    f"Missing argument for {flash_method}: 'calculated_variable'"
                )

        if flash_method == "isenthalpic":
            enth_input = {
                "unit": self.input_unit_set["enthalpy"]["oli_unit"],
                "value": None,
            }
            if float(enthalpy):
                enth_input.update({"value": float(enthalpy)})
            else:
                raise ValueError(f"Invalid enthalpy: {enthalpy}. Expected number")
            input_dict["enthalpy"] = enth_input

        if flash_method == "vapor-amount":
            vapor_amount_input = (
                {
                    "unit": input_unit_set["vaporAmountMoles"]["oli_unit"],
                    "value": None,
                },
            )
            if float(vapor_amount):
                vapor_amount_input.update({"value": float(vapor_amount)})
            else:
                raise ValueError(
                    f"Invalid vapor amount: {vapor_amount}. Expected number"
                )
            input_dict["vaporAmountMoles"] = vapor_amount_input

        if flash_method == "vapor-fraction":
            vapor_fraction_amount = (
                {
                    "unit": input_unit_set["vaporMolFrac"]["oli_unit"],
                    "value": None,
                },
            )
            if float(vapor_fraction):
                vapor_fraction_amount.update({"value": float(vapor_fraction)})
            else:
                raise ValueError(
                    f"Invalid vapor fraction: {vapor_fraction}. Expected number"
                )
            input_dict["vaporMolFrac"] = vapor_fraction_amount

        if flash_method == "isochoric":
            volume_input = {
                "unit": input_unit_set["totalVolume"]["oli_unit"],
                "value": None,
            }
            if float(volume):
                volume_input.update({"value": float(volume)})
            else:
                raise ValueError(f"Invalid volume: {volume}. Expected number")
            input_dict["totalVolume"] = volume_input

        if flash_method == "setph":
            ph_input = {
                "targetPH": {
                    "unit": "",
                    "value": None,
                },
            }
            if float(ph):
                ph_input["targetPH"].update({"value": float(ph)})
            else:
                raise ValueError(f"Invalid ph: {ph}. Expected number")
            input_dict["targetPH"] = ph_input
            if not acid_titrant:
                acid_titrant = "HCl"
            input_dict["pHAcidTitrant"] = get_oli_name(acid_titrant)
            if not base_titrant:
                base_titrant = "NaOH"
            input_dict["pHBaseTitrant"] = get_oli_name(base_titrant)

        if flash_method == "precipitation-point":
            missing_args = [
                arg for arg in [formed_solid, precipitant_inflow] if arg is None
            ]
            if missing_args:
                raise RuntimeError(
                    f"Missing argument(s) for {flash_method}: {missing_args}"
                )
            else:
                input_dict.update(
                    {
                        "solidToPrecipitate": formed_solid,
                        "inflowToAdjust": precipitant_inflow,
                    }
                )

        if ("unit" in inflows.keys()) and ("values" in inflows.keys()):
            input_dict["inflows"] = inflows
        # otherwise, assume a solute dictionary with {solute_name: concentration_value}
        else:
            unit = self.input_unit_set["molecularConcentration"]["oli_unit"]
            inflows_oli_names = {get_oli_name(k): v for k, v in inflows.items()}
            input_dict["inflows"] = {"unit": unit, "values": inflows_oli_names}

        if flash_method == "corrosion-rates":
            _logger.info(
                f"Ensure DBS file uses 'AQ' thermodynamic framework to use Corrosion Analyzer"
            )
            input_dict["corrosionParameters"] = self._configure_corrosion(
                contact_surface,
                flow_type,
                diameter,
                liq_velocity,
                gas_velocity,
                rot_velocity,
                shear_stress,
                roughness,
                nonaqueous_visc,
                water_cut_inversion,
                relative_visc_inversion,
            )

        json_input = self._add_to_json(
            flash_method,
            input_dict,
            included_solids,
            excluded_solids,
            use_scaling_rigorous,
            file_name,
        )
        return json_input

    def _configure_corrosion(
        self,
        contact_surface,
        flow_type,
        diameter,
        liq_velocity,
        gas_velocity,
        rot_velocity,
        shear_stress,
        roughness,
        nonaqueous_visc,
        water_cut_inversion,
        relative_visc_inversion,
    ):
        """
        Configure input dict for Corrosion Rates flash.

        :param contact_surface: string for contact surface metal name
        :param flow_type: string for flow configuration
        :param diameter: float for diameter of surface (i.e., pipe or rotor)
        :param liq_velocity: float for velocity of liquid flow
        :param gas_velocity: float for velocity of vapor flow, used in 'approximateMultiPhaseFlow'
        :param rot_velocity: float for rotational velocity
        :param shear_stress: float for defined shear stress, used in 'definedShearStress'
        :param roughness: float for pipe roughness, used in 'approximateMultiPhaseFlow'
        :param nonaqueous_visc: float for absolute viscosity of nonaqueous phase, used in 'approximateMultiPhaseFlow'
        :param water_cut_inversion: float for water cut at point of dispersion inversion, used in 'approximateMultiPhaseFlow'
        :param relative_visc_inversion: float for maximum relative viscosity of dispersion at inversion, used in 'approximateMultiPhaseFlow'

        :return config: dictionary for corrosion analysis parameters
        """

        valid_flow_types = [
            "static",
            "pipeFlow",
            "rotatingDisk",
            "rotatingCylinder",
            "completeAgitation",
            "definedShearStress",
            "approximateMultiPhaseFlow",
        ]
        if flow_type not in valid_flow_types:
            raise RuntimeError(
                f"Invalid flow_type: {flow_type}."
                f"Expected one of {', '.join(t for t in valid_flow_types)}"
            )

        config = {
            "calculationType": "isothermal",
            "corrosionParameters": {
                "contactSurface": contact_surface,
                "flowType": flow_type,
            },
        }

        def _try_float(v):
            try:
                val = float(v)
            except:
                val = None
            return val

        _check_args = lambda args: [arg for arg in args if arg is None]
        _check_floats = lambda args: [arg for arg in args if _try_float(arg) is None]
        if flow_type == "pipeFlow":
            args = [diameter, liq_velocity]
            missing_args = _check_args(args)
            not_floats = _check_floats(args)
        elif flow_type in ["rotatingDisk", "rotatingCylinder"]:
            args = [diameter, rot_velocity]
            missing_args = _check_args(args)
            not_floats = _check_floats(args)
        elif flow_type == "definedShearStress":
            args = [shear_stress]
            missing_args = _check_args(args)
            not_floats = _check_floats(args)
        elif flow_type == "approximateMultiPhaseFlow":
            args = [
                diameter,
                liq_velocity,
                gas_velocity,
                roughness,
                nonaqueous_visc,
                water_cut_inversion,
                relative_visc_inversion,
            ]
            missing_args = _check_args(args)
            not_floats = _check_floats(args)
        if missing_args:
            raise RuntimeError(
                f"Missing argument(s) for {flash_method}: {missing_args}"
            )
        if not_floats:
            raise RuntimeError(
                f"Invalid values for argument(s): {not_floats}. Expected value"
            )

        if flow_type == "pipeFlow":
            config["corrosionParameters"].update(
                {
                    "pipeDiameter": {
                        "value": diameter,
                        "unit": self.input_unit_set["pipeDiameter"]["oli_unit"],
                    },
                    "pipeFlowVelocity": {
                        "value": liq_velocity,
                        "unit": self.input_unit_set["pipeFlowVelocity"]["oli_unit"],
                    },
                }
            )
        elif flow_type == "rotatingDisk":
            config["corrosionParameters"].update(
                {
                    "diskDiameter": {
                        "value": diameter,
                        "unit": self.input_unit_set["diskDiameter"]["oli_unit"],
                    },
                    "diskRotationSpeed": {
                        "value": rot_velocity,
                        "unit": self.input_unit_set["diskRotationSpeed"]["oli_unit"],
                    },
                },
            )

        elif flow_type == "rotatingCylinder":
            config["corrosionParameters"].update(
                {
                    "rotorDiameter": {
                        "value": diameter,
                        "unit": self.input_unit_set["rotorDiameter"]["oli_unit"],
                    },
                    "rotorRotation": {
                        "value": rot_velocity,
                        "unit": self.input_unit_set["rotorRotation"]["oli_unit"],
                    },
                },
            )

        elif flow_type == "definedShearStress":
            config["corrosionParameters"].update(
                {
                    "shearStress": {
                        "value": shear_stress,
                        "unit": self.input_unit_set["shearStress"]["oli_unit"],
                    },
                },
            )

        elif flow_type == "approximateMultiPhaseFlow":
            config["corrosionParameters"].update(
                {
                    "pipeDiameter": {
                        "value": diameter,
                        "unit": self.input_unit_set["pipeDiameter"]["oli_unit"],
                    },
                    "liquidFlowInPipe": {
                        "value": liq_velocity,
                        "unit": self.input_unit_set["liquidFlowInPipe"]["oli_unit"],
                    },
                    "gasFlowInPipe": {
                        "value": gas_velocity,
                        "unit": self.input_unit_set["gasFlowInPipe"]["oli_unit"],
                    },
                    "pipeRoughness": {
                        "value": roughness,
                        "unit": self.input_unit_set["pipeRoughness"]["oli_unit"],
                    },
                    "viscAbs2ndLiq": {
                        "value": nonaqueous_visc,
                        "unit": self.input_unit_set["viscAbs2ndLiq"]["oli_unit"],
                    },
                    "waterCutAtPointOfDispersionInversion": water_cut_inversion,
                    "maxRelViscosityOfDispersionAtInversion": relative_visc_inversion,
                }
            )
        return config

    def _add_to_json(
        self,
        flash_method,
        input_data,
        included_solids,
        excluded_solids,
        use_scaling_rigorous,
        file_name,
    ):
        """
        Add input data to JSON.

        :param flash_method: string for flash calculation name
        :param input_data: data object from flash configuration function
        :param included_solids: list of solids to include in analysis
        :param excluded_solids: list of solids to exclude from analysis
        :param use_scaling_rigorous: bool to switch between Rigorous (default) and Estimated scaling computations
        :param file_name: string for file to write, if any
        """

        self._set_prescaling_calculation_mode(use_scaling_rigorous)
        json_input = {"params": {}}

        if flash_method == "wateranalysis":
            json_input["params"].update({"waterAnalysisInputs": input_data})
        else:
            json_input["params"] = input_data

        output_unit_set_info = {}
        for k, v in self.output_unit_set.items():
            output_unit_set_info[k] = v["oli_unit"]
        additional_params = {
            "optionalProperties": dict(self.optional_properties),
            "unitSetInfo": dict(output_unit_set_info),
        }
        if included_solids and excluded_solids:
            raise RuntimeError(
                "Invalid argument combination. "
                "Only one of included_solids and excluded_solids "
                "may be specified at once."
            )
        else:
            if included_solids:
                additional_params.update({"included_solids": list(included_solids)})
            if excluded_solids:
                additional_params.update({"excluded_solids": list(excluded_solids)})
        json_input["params"].update(additional_params)

        if file_name is not None:
            write_output(json_input, file_name)
        return json_input

    def _set_prescaling_calculation_mode(self, use_scaling_rigorous):
        """
        Set prescaling computation method based on argument.

        :param use_scaling_rigorous: boolean indicating desired state of 'rigorous' and 'estimated' optional properties
        """

        props = self.optional_properties
        if bool(use_scaling_rigorous) == bool(props["prescalingTendenciesRigorous"]):
            return
        new_values = {k: not v for k, v in props.items() if "prescaling" in k}
        props.update(new_values)

    def run_flash(
        self,
        flash_method,
        oliapi_instance,
        dbs_file_id,
        json_input,
        survey=None,
        file_name=None,
        # max_concurrent_processes=1000,
        # burst_job_tag=None,
        # batch_size=None,
    ):
        """
        Conduct single point analysis with initial JSON input, or conduct a survey on that input.

        :param flash_method: string for flash calculation name
        :param oliapi_instance: instance of OLI Cloud API
        :param dbs_file_id: string ID of DBS file
        :param json_input: JSON input for flash calculation
        :param survey: dictionary containing names and input values to modify in JSON
        :param file_name: string for file to write, if any

        :return processed_requests: results from processed OLI flash requests
        """

        if flash_method == "corrosion-rates":
            # check if DBS file is using AQ thermodynamic framework
            oliapi_instance.get_corrosion_contact_surfaces(dbs_file_id)

        if self.relative_inflows:
            _logger.info(
                f"relative_inflows={self.relative_inflows},"
                + " surveys will add values to initial state"
            )

        if survey is None:
            survey = {}
        num_samples = None
        for k, v in survey.items():
            if num_samples is None:
                num_samples = len(v)
            elif num_samples != len(v):
                raise RuntimeError(f"Length of list for key {k} differs from prior key")
        if num_samples is None:
            num_samples = 1
        requests_to_process = []
        for idx in range(num_samples):
            _logger.info(f"Flash sample #{idx+1} of {num_samples}")
            requests_to_process.append(
                {
                    "flash_method": flash_method,
                    "dbs_file_id": dbs_file_id,
                    "input_params": self.get_clone(
                        flash_method, json_input, idx, survey
                    ),
                }
            )
        processed_requests = oliapi_instance.process_request_list(
            requests_to_process,
            # burst_job_tag=burst_job_tag,
            # max_concurrent_processes=max_concurrent_processes,
            # batch_size=batch_size,
        )
        _logger.info("Completed running flash calculations")
        result = flatten_results(processed_requests)
        if file_name:
            write_output(result, file_name)
        return result

    def get_clone(self, flash_method, json_input, index, survey=None):
        """
        Iterate over a survey to create a modified clone from JSON input.

        :param flash_method: string for flash calculation name
        :param json_input: JSON input for flash calculation
        :param index: integer for index of incoming data
        :param survey: dictionary containing names and input values to modify in JSON

        :return clone: dictionary containing modified state variables and survey index
        """

        if survey is None:
            return json_input

        valid_flashes = [
            "wateranalysis",
            "isothermal",
            "isenthalpic",
            "bubblepoint",
            "dewpoint",
            "vapor-amount",
            "vapor-fraction",
            "isochoric",
            "setph",
            "precipitation-point",
            "corrosion-rates",
        ]
        if flash_method not in valid_flashes:
            raise RuntimeError(
                "Invalid flash_method: {flash_method}. Use one of {', '.join(valid_flashes)}"
            )

        clone = deepcopy(json_input)
        for k, v in survey.items():
            d = clone["params"]
            if flash_method == "wateranalysis":
                d = d["waterAnalysisInputs"]
                for param in d:
                    if param["name"].lower() == k.lower():
                        if self.relative_inflows:
                            param["value"] += v[index]
                        else:
                            param["value"] = v[index]
                        _logger.info(
                            f"Updating {k} for sample #{index} clone: new value = {param['value']}"
                        )
            else:
                if k in d:
                    pass
                elif k in d["inflows"]["values"]:
                    d = d["inflows"]["values"]
                elif hasattr(d, "corrosionParameters") and (
                    k in d["corrosionParameters"]
                ):
                    d = d["corrosionParameters"]
                else:
                    _logger.warning(f"Survey key {k} not found in JSON input.")
                if self.relative_inflows:
                    if isinstance(d[k], dict):
                        d[k]["value"] += v[index]
                        val = d[k]["value"]
                    else:
                        d[k] += v[index]
                        val = d[k]
                else:
                    if isinstance(d[k], dict):
                        d[k]["value"] = v[index]
                        val = d[k]["value"]
                    else:
                        d[k] = v[index]
                        val = d[k]
                _logger.info(
                    f"Updating {k} for sample #{index} clone: new value = {val}"
                )
        return clone

    def get_apparent_species_from_true(
        self,
        true_species_json,
        oliapi_instance,
        dbs_file_id,
        phase=None,
        file_name=None,
    ):
        """
        Run Water Analysis to get apparent species.

        :param true_species_json: JSON generated from true species
        :param oliapi_instance: instance of OLI Cloud API to call
        :param dbs_file_id: string ID of DBS file
        :param phase: string for inflows phase
        :param file_name: string for file to write, if any

        :return apparent_species: dictionary for molecular concentrations
        """

        stream_output = self.run_flash(
            "wateranalysis",
            oliapi_instance,
            dbs_file_id,
            true_species_json,
        )
        if phase is None:
            phase = "total"

        extracted_result = stream_output["result"][f"molecularConcentration_{phase}"]

        unit = self.input_unit_set["molecularConcentration"]["oli_unit"]
        concentrations = {k: v["values"][0] for k, v in extracted_result.items()}
        inflows = {"unit": unit, "values": concentrations}
        if file_name:
            write_output(inflows, file_name)
        return inflows


def flatten_results(processed_requests):

    _logger.info("Flattening OLI stream output ... ")

    props = []
    terminal_keys = ["unit", "value", "found", "fullVersion", "values"]

    def _find_props(data, path=None):
        """
        Get the path to all nested items in input data (recursive search).

        :param data: dictionary containing OLI flash output
        :param path: list of paths to endpoint

        :return props: list of nested path lists
        """
        path = path if path is not None else []
        if isinstance(data, dict):
            for k, v in data.items():
                if isinstance(v, (str, bool)):
                    props.append([*path, k])
                elif isinstance(v, list):
                    if all(k not in terminal_keys for k in v):
                        _find_props(v, [*path, k])
                elif isinstance(v, dict):
                    if all(k not in terminal_keys for k in v):
                        _find_props(v, [*path, k])
                    else:
                        props.append([*path, k])
        elif isinstance(data, list):
            for idx, v in enumerate(data):
                if isinstance(v, (dict, list)):
                    if all(k not in terminal_keys for k in v):
                        _find_props(v, [*path, idx])
                    else:
                        props.append([*path, idx])
        else:
            raise RuntimeError(f"Unexpected type for data: {type(data)}")

    def _get_nested_data(data, keys):

        for key in keys:
            data = data[key]
        return data

    def _extract_values(data, keys):
        values = _get_nested_data(data, keys)
        extracted_values = {}
        if isinstance(values, str):
            extracted_values = values
        elif isinstance(values, bool):
            extracted_values = bool(values)
        elif isinstance(values, dict):
            if any(k in values for k in ["group", "name", "fullVersion"]):
                if "value" in values:
                    extracted_values.update({"values": values["value"]})
                if "unit" in values:
                    unit = values["unit"] if values["unit"] else "dimensionless"
                    extracted_values.update({"units": unit})

            elif all(k in values for k in ["found", "phase"]):
                extracted_values = values
            else:
                unit = values["unit"] if values["unit"] else "dimensionless"
                if "value" in values:
                    extracted_values = {
                        "units": unit,
                        "values": values["value"],
                    }
                elif "values" in values:
                    extracted_values = {
                        k: {
                            "units": unit,
                            "values": values["values"][k],
                        }
                        for k, v in values["values"].items()
                    }
                elif "data" in values:
                    # intended for vaporDiffusivityMatrix
                    mat_dim = int(sqrt(len(values["data"])))
                    diffmat = reshape(values["data"], newshape=(mat_dim, mat_dim))

                    extracted_values = {
                        f'({values["speciesNames"][i]},{values["speciesNames"][j]})': {
                            "units": values["unit"],
                            "values": diffmat[i][j],
                        }
                        for i in range(len(diffmat))
                        for j in range(i, len(diffmat))
                    }
                else:
                    raise NotImplementedError(
                        f"results structure not accounted for. results:\n{values}"
                    )
        else:
            raise RuntimeError(f"Unexpected type for data: {type(values)}")
        return extracted_values

    def _create_input_dict(props, result):
        input_dict = {k: {} for k in set([prop[0] for prop in props])}
        for prop in props:
            k = prop[0]
            phase_tag = ""
            if "metaData" in prop:
                prop_tag = prop[-1]
            elif "result" in prop:
                # get property tag
                if isinstance(prop[-1], int):
                    prop_tag = prop[-2]
                else:
                    prop_tag = prop[-1]
                # get phase tag
                if any(k in prop for k in ["phases", "total"]):
                    if "total" in prop:
                        phase_tag = "total"
                    else:
                        phase_tag = prop[prop.index("phases") + 1]
            elif "submitted_requests" in prop:
                prop_tag = prop[-1]
                if "params" in prop:
                    if isinstance(prop[-1], int):
                        prop_tag = _get_nested_data(result, prop)["name"]
            else:
                _logger.warning(
                    f"Unexpected result:\n{result}\n\ninput_dict:\n{input_dict} from prop {prop}"
                )
                continue
            label = f"{prop_tag}_{phase_tag}" if phase_tag else prop_tag
            input_dict[k][label] = _extract_values(result, prop)
        return input_dict

    float_nan = float("nan")

    def _add_to_output(input_dict, output_dict, index, number_samples):
        """
        Add incoming flash results to output data.

        :param input_dict: dictionary for incoming data
        :param output_dict: dictionary for output data
        :param index: integer for index of incoming data
        :param number_samples: integer for total number of incoming data samples
        """

        for k, v in input_dict.items():
            try:
                val = float(v)
            except:
                val = None
            if val is not None:
                if k not in output_dict:
                    output_dict[k] = [float_nan] * number_samples
                output_dict[k][index] = val
            elif isinstance(v, str):
                if k not in output_dict:
                    if k in ["fullVersion", "units"]:
                        output_dict[k] = v
                    else:
                        output_dict[k] = [float_nan] * number_samples
                if k in ["fullVersion", "units"]:
                    if input_dict[k] != output_dict[k]:
                        raise Exception(f"Input and output do not agree for key {k}")
                else:
                    output_dict[k][index] = v
            elif isinstance(v, dict):
                if k not in output_dict:
                    output_dict[k] = {}
                _add_to_output(input_dict[k], output_dict[k], index, number_samples)
            else:
                raise Exception(f"Unexpected value: {v}")

    output_dict = {}
    num_samples = len(processed_requests)
    for idx, result in enumerate(processed_requests):
        props = []
        _find_props(result)
        input_dict = _create_input_dict(props, result)
        _add_to_output(input_dict, output_dict, idx, num_samples)
    return output_dict


def write_output(content, file_name):
    """
    Write dictionary-based content to file.

    :param content: dictionary of content to write
    :param file_name: string for name of file to write

    :param file_path: string for full path of written file
    """

    _logger.info(f"Saving content to {file_name}")
    with open(file_name, "w", encoding="utf-8") as f:
        json.dump(content, f)
    _logger.info("Save complete")
    file_path = Path(file_name).resolve()
    return file_path


def build_survey(survey_arrays, get_oli_names=False, file_name=None, mesh_grid=True):
    """
    Build a dictionary for modifying flash calculation parameters.

    :param survey_arrays: dictionary for variables and values to survey
    :param get_oli_names: bool switch to convert name into OLI name
    :param file_name: string for file to write, if any
    :param mesh_grid: if True (default) the input array will be combined to generate combination of all possible samples
        if False, the direct values in survey_arrays will be used

    :return survey: dictionary for product of survey variables and values
    """
    _name = lambda k: get_oli_name(k) if get_oli_names else k
    if mesh_grid:
        keys = [get_oli_name(k) if get_oli_names else k for k in survey_arrays]
        values = list(product(*(survey_arrays.values())))
        survey = {_name(keys[i]): [val[i] for val in values] for i in range(len(keys))}
    else:
        survey = {}
        values = None
        for key, arr in survey_arrays.items():
            survey[_name(key)] = arr
            if values is not None and len(values) != len(arr):
                raise ValueError(f"Length of list for key {key} differs from prior key")
            values = arr
    _logger.info(f"Survey contains {len(values)} items.")
    if file_name:
        write_output(survey, file_name)
    return survey


def get_survey_sample_conditions(survey, sample_points):
    """
    Return survey parameter values for one or more sample points.

    :param survey: dictionary for product of survey conditions and values
    :param sample_points: list of indices to get parameter values from

    :return sample_conditions: dictionary for parameter values for given samples
    """

    sample_conditions = {}
    for point in sample_points:
        sample_conditions[point] = {}
        for k, v in survey.items():
            sample_conditions[point][k] = v[point]
    _logger.debug(sample_conditions)
    return sample_conditions



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo

from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.core.expr.visitor import identify_variables

from idaes.core.base.costing_base import register_idaes_currency_units
from idaes.core import declare_process_block_class, UnitModelBlockData
from idaes.core.base.costing_base import FlowsheetCostingBlockData
from idaes.models.unit_models import Mixer, HeatExchanger, Heater, CSTR

import idaes.logger as idaeslog

from watertap.core.util.misc import is_constant_up_to_units
from watertap.costing.unit_models.mixer import cost_mixer
from watertap.costing.unit_models.heat_exchanger import cost_heat_exchanger
from watertap.costing.unit_models.cstr import cost_cstr
from watertap.costing.unit_models.heater_chiller import cost_heater_chiller

_log = idaeslog.getLogger(__name__)


class WaterTAPCostingBlockData(FlowsheetCostingBlockData):
    """
    Base class for creating WaterTAP costing packages. Allows
    unit models to "self-register" their default costing methods,
    and for anonymous expressions in flow costs.
    """

    # Define default mapping of costing methods to unit models
    unit_mapping = {
        Mixer: cost_mixer,
        HeatExchanger: cost_heat_exchanger,
        CSTR: cost_cstr,
        Heater: cost_heater_chiller,
    }

    def register_currency_definitions(self):
        # Register currency and conversion rates based on CE Index
        register_idaes_currency_units()

    def set_base_currency_base_period(self):
        # Set the base year for all costs
        self.base_currency = pyo.units.USD_2018
        # Set a base period for all operating costs
        self.base_period = pyo.units.year

    def add_LCOW(self, flow_rate, name="LCOW"):
        """
        Add Levelized Cost of Water (LCOW) to costing block.
        Args:
            flow_rate - flow rate of water (volumetric) to be used in
                        calculating LCOW
            name (optional) - name for the LCOW variable (default: LCOW)
        """

        denominator = (
            pyo.units.convert(flow_rate, to_units=pyo.units.m**3 / self.base_period)
            * self.utilization_factor
        )

        self.add_component(
            name,
            pyo.Expression(
                expr=(
                    self.total_capital_cost * self.capital_recovery_factor
                    + self.total_operating_cost
                )
                / denominator,
                doc=f"Levelized Cost of Water based on flow {flow_rate.name}",
            ),
        )

        c_units = self.base_currency
        t_units = self.base_period
        direct_capex_lcows = pyo.Expression(
            pyo.Any,
            doc=f"Levelized Cost of Water based on flow {flow_rate.name} direct capital expenditure by component",
            initialize=0 * c_units / pyo.units.m**3,
        )
        indirect_capex_lcows = pyo.Expression(
            pyo.Any,
            doc=f"Levelized Cost of Water based on flow {flow_rate.name} indirect capital expenditure by component",
            initialize=0 * c_units / pyo.units.m**3,
        )
        fixed_opex_lcows = pyo.Expression(
            pyo.Any,
            doc=f"Levelized Cost of Water based on flow {flow_rate.name} fixed operating expenditure by component",
            initialize=0 * c_units / pyo.units.m**3,
        )
        variable_opex_lcows = pyo.Expression(
            pyo.Any,
            doc=f"Levelized Cost of Water based on flow {flow_rate.name} variable operating expenditure by component",
            initialize=0 * c_units / pyo.units.m**3,
        )
        self.add_component(name + "_component_direct_capex", direct_capex_lcows)
        self.add_component(name + "_component_indirect_capex", indirect_capex_lcows)
        self.add_component(name + "_component_fixed_opex", fixed_opex_lcows)
        self.add_component(name + "_component_variable_opex", variable_opex_lcows)

        agg_direct_capex_lcows = pyo.Expression(
            pyo.Any,
            doc=f"Levelized Cost of Water based on flow {flow_rate.name} direct capital expenditure by unit type",
            initialize=0 * c_units / pyo.units.m**3,
        )
        agg_indirect_capex_lcows = pyo.Expression(
            pyo.Any,
            doc=f"Levelized Cost of Water based on flow {flow_rate.name} indirect capital expenditure by unit type",
            initialize=0 * c_units / pyo.units.m**3,
        )
        agg_fixed_opex_lcows = pyo.Expression(
            pyo.Any,
            doc=f"Levelized Cost of Water based on flow {flow_rate.name} fixed operating expenditure by unit type",
            initialize=0 * c_units / pyo.units.m**3,
        )
        agg_variable_opex_lcows = pyo.Expression(
            pyo.Any,
            doc=f"Levelized Cost of Water based on flow {flow_rate.name} variable operating expenditure by unit type",
            initialize=0 * c_units / pyo.units.m**3,
        )
        self.add_component(name + "_aggregate_direct_capex", agg_direct_capex_lcows)
        self.add_component(name + "_aggregate_indirect_capex", agg_indirect_capex_lcows)
        self.add_component(name + "_aggregate_fixed_opex", agg_fixed_opex_lcows)
        self.add_component(name + "_aggregate_variable_opex", agg_variable_opex_lcows)
        for u in self._registered_unit_costing:
            direct_capex_numerator = 0 * c_units / t_units
            indirect_capex_numerator = 0 * c_units / t_units
            fixed_opex_numerator = 0 * c_units / t_units
            variable_opex_numerator = 0 * c_units / t_units
            if hasattr(u, "capital_cost"):
                direct_capital_cost = pyo.units.convert(
                    u.direct_capital_cost, to_units=c_units
                )
                capital_cost = pyo.units.convert(u.capital_cost, to_units=c_units)
                # capital costs w/ recovery factor
                direct_capex_numerator += (
                    self.capital_recovery_factor * direct_capital_cost
                )
                capex_numerator = self.capital_recovery_factor * (
                    self.total_investment_factor * capital_cost
                )
                indirect_capex_numerator += capex_numerator - direct_capex_numerator

                # maintenance_labor_chemical_operating_cost,
                # part of total_fixed_operating_cost
                fixed_opex_numerator += (
                    self.maintenance_labor_chemical_factor * capital_cost
                )
            if hasattr(u, "fixed_operating_cost"):
                fixed_opex_numerator += pyo.units.convert(
                    u.fixed_operating_cost, to_units=c_units / t_units
                )
            if hasattr(u, "variable_operating_cost"):
                variable_opex_numerator += pyo.units.convert(
                    u.variable_operating_cost, to_units=c_units / t_units
                )
            direct_capex_lcows[u.unit_model.name] = direct_capex_numerator / denominator
            indirect_capex_lcows[u.unit_model.name] = (
                indirect_capex_numerator / denominator
            )
            fixed_opex_lcows[u.unit_model.name] = fixed_opex_numerator / denominator
            variable_opex_lcows[u.unit_model.name] = (
                variable_opex_numerator / denominator
            )

            unit_model_class = (
                u.unit_model.parent_component().process_block_class().__name__
            )
            agg_direct_capex_lcows[unit_model_class] += direct_capex_lcows[
                u.unit_model.name
            ]
            agg_indirect_capex_lcows[unit_model_class] += indirect_capex_lcows[
                u.unit_model.name
            ]
            agg_fixed_opex_lcows[unit_model_class] += fixed_opex_lcows[
                u.unit_model.name
            ]
            agg_variable_opex_lcows[unit_model_class] += variable_opex_lcows[
                u.unit_model.name
            ]

        for ftype, flows in self._registered_flows.items():
            # part of total_variable_operating_cost
            if ftype in agg_variable_opex_lcows:
                raise RuntimeError(
                    f"Found unit model named {ftype} but want to name flow {ftype} for {name+'_component_variable_opex'}"
                )
            if ftype not in self.used_flows:
                assert len(flows) == 0
                continue
            agg_variable_opex_lcows[ftype] = (
                self.aggregate_flow_costs[ftype] * self.utilization_factor / denominator
            )
            cost_var = getattr(self, f"{ftype}_cost")
            for flow_expr in flows:
                # part of total_variable_operating_cost
                flow_cost = pyo.units.convert(
                    flow_expr * cost_var, to_units=c_units / t_units
                )
                unit = self._find_flow_unit(flow_expr)
                if unit is not None:
                    if unit.name not in variable_opex_lcows:
                        unit_model_class = (
                            unit.parent_component().process_block_class().__name__
                        )
                        agg_variable_opex_lcows[
                            unit_model_class
                        ] += variable_opex_lcows[unit.name]
                    variable_opex_lcows[unit.name] += (
                        flow_cost * self.utilization_factor
                    ) / denominator
                    continue
                _log.warning(f"Could not find unique unit for flow {flow_expr}")
                flow_name = self._get_flow_name(flow_expr)
                if flow_name in variable_opex_lcows:
                    raise RuntimeError(
                        f"Found unit model named {flow_name} but need to name flow {flow_expr} for {name+'_aggregate_variable_opex'}"
                    )
                variable_opex_lcows[flow_name] = (
                    flow_cost * self.utilization_factor
                ) / denominator

    @staticmethod
    def _find_flow_unit(flow_expr):
        """
        Attempts to identify a unit model associated with the flow expression
        `flow_expr`. Returns the unit model if found, otherwise returns `None`.
        """
        variables = identify_variables(flow_expr, include_fixed=True)
        unit = WaterTAPCostingBlockData._find_flow_unit_variables(variables)
        return unit

    @staticmethod
    def _find_flow_unit_variables(variables):
        """
        Try to find a unit model uniquely associated with the iterable
        of variables `variables`.
        """
        unit = None
        for v in variables:
            blk = v.parent_block()
            while blk is not None:
                if UnitModelBlockData in blk.__class__.__mro__:
                    if unit is None:
                        unit = blk
                        break
                    elif blk is unit:
                        break
                    else:
                        # blk is unit model but not *this* unit,
                        # so we have an inconsistency
                        return None
                blk = blk.parent_block()
        return unit

    @staticmethod
    def _get_flow_name(flow_expr):
        """
        Get a standardized name for a specific flow expression.

        This will attempt to get the specific variable determining the flow,
        but failing that will return a string representation of the entire
        expression.
        """
        variables = list(identify_variables(flow_expr, include_fixed=False))
        if not variables:
            # No unfixed variables
            variables = list(identify_variables(flow_expr, include_fixed=True))
        # If we found a single variable for the flow_expr, just return the
        # variable name. Otherwise return a string representing the whole
        # expression.
        if len(variables) == 1:
            return str(variables[0])
        return str(flow_expr)

    def add_specific_energy_consumption(
        self, flow_rate, name="specific_energy_consumption"
    ):
        """
        Add specific energy consumption (kWh/m**3) to costing block.
        Args:
            flow_rate - flow rate of water (volumetric) to be used in
                        calculating specific energy consumption
            name (optional) - the name of the Expression for the specific
                              energy consumption (default: specific_energy_consumption)
        """

        self.add_component(
            name,
            pyo.Expression(
                expr=self.aggregate_flow_electricity
                / pyo.units.convert(flow_rate, to_units=pyo.units.m**3 / pyo.units.hr),
                doc=f"Specific energy consumption based on flow {flow_rate.name}",
            ),
        )
        self._add_flow_component_breakdowns(
            "electricity", name, flow_rate, utilization_factor=1.0, period=pyo.units.hr
        )

    def add_annual_water_production(self, flow_rate, name="annual_water_production"):
        """
        Add annual water production to costing block.
        Args:
            flow_rate - flow rate of water (volumetric) to be used in
                        calculating annual water production
            name (optional) - name for the annual water productionvariable
                              Expression (default: annual_water_production)
        """
        self.add_component(
            name,
            pyo.Expression(
                expr=(
                    pyo.units.convert(
                        flow_rate, to_units=pyo.units.m**3 / self.base_period
                    )
                    * self.utilization_factor
                ),
                doc=f"Annual water production based on flow {flow_rate.name}",
            ),
        )

    def add_electricity_intensity(self, flow_rate, name="electricity_intensity"):
        """
        Add calculation of overall electricity intensity to costing block.
        Args:
            flow_rate - flow rate of water (volumetric) to be used in
                        calculating electricity intensity
            name (optional) - the name of the Expression for the specific
                              electrical intensity (default: specific_electrical_carbon_intensity)
        """
        self.add_specific_energy_consumption(flow_rate, name=name)

    def add_specific_electrical_carbon_intensity(
        self, flow_rate, name="specific_electrical_carbon_intensity"
    ):
        """
        Add specific electrical carbon intensity (kg_CO2eq/m**3) to costing block.
        Args:
            flow_rate - flow rate of water (volumetric) to be used in
                        calculating specific electrical carbon intensity
            name (optional) - the name of the Expression for the specific
                              carbon intensity (default: specific_electrical_carbon_intensity)
        """

        self.add_component(
            name,
            pyo.Expression(
                expr=self.aggregate_flow_electricity
                * self.electrical_carbon_intensity
                / pyo.units.convert(flow_rate, to_units=pyo.units.m**3 / pyo.units.hr),
                doc=f"Specific electrical carbon intensity based on flow {flow_rate.name}",
            ),
        )
        self._add_flow_component_breakdowns(
            "electricity",
            name,
            flow_rate,
            period=pyo.units.hr,
            utilization_factor=1.0,
            multiplier=self.electrical_carbon_intensity,
        )

    def _add_flow_component_breakdowns(
        self,
        flow_name,
        name,
        flow_rate,
        period=None,
        utilization_factor=None,
        multiplier=1.0,
    ):
        """
        Add per-component breakdowns for specific `flow_name` consumption with base-name `name`
        at `flow_rate`.
        Optional `multiplier` for the flow and period specification (default is 1 hour),
        and specified `utilization_factor` (default is self.utilization_factor).
        """
        if utilization_factor is None:
            utilization_factor = self.utilization_factor
        if period is None:
            period = self.base_period
        denominator = (
            pyo.units.convert(flow_rate, to_units=pyo.units.m**3 / period)
            * utilization_factor
        )
        f_units = pyo.units.get_units(getattr(self, f"aggregate_flow_{flow_name}"))
        c_units = f_units * pyo.units.get_units(multiplier)

        try:
            flows = self._registered_flows[flow_name]
        except KeyError:
            raise RuntimeError(f"Unrecognized flow_name {flow_name}.")

        specific_flow_consumption = pyo.Expression(
            pyo.Any,
            doc=f"Specific {flow_name} consumption by component",
            initialize=0.0 * period * c_units / pyo.units.m**3,
        )
        self.add_component(name + "_component", specific_flow_consumption)

        for flow_expr in flows:
            flow_std = pyo.units.convert(flow_expr, to_units=f_units)
            unit = self._find_flow_unit(flow_expr)
            if unit is not None:
                specific_flow_consumption[unit.name] += (
                    flow_std * utilization_factor * multiplier
                ) / denominator
                continue
            _log.warning(f"Could not find unique unit for flow {flow_expr}")
            flow_name = self._get_flow_name(flow_expr)
            specific_flow_consumption[flow_name] += (
                flow_std * utilization_factor * multiplier
            ) / denominator

    def build_process_costs(self):
        """
        Build the common process costs to WaterTAP Costing Packages.
        The currency units should already be registered.

        The derived class should add constraints for total_capital_cost
        and total_operating_cost
        """
        self.total_capital_cost = pyo.Var(
            initialize=0,
            doc="Total capital cost of the process",
            units=self.base_currency,
        )
        self.total_operating_cost = pyo.Var(
            initialize=0,
            doc="Total operating cost of process per operating period",
            units=self.base_currency / self.base_period,
        )

        self.total_capital_cost_constraint = pyo.Constraint(
            expr=self.total_capital_cost
            == self.total_investment_factor * self.aggregate_capital_cost
        )

        self.maintenance_labor_chemical_operating_cost = pyo.Expression(
            expr=self.maintenance_labor_chemical_factor * self.aggregate_capital_cost,
            doc="Maintenance-labor-chemical operating cost",
        )

        self.total_fixed_operating_cost = pyo.Expression(
            expr=self.aggregate_fixed_operating_cost
            + self.maintenance_labor_chemical_operating_cost,
            doc="Total fixed operating costs",
        )

        self.total_variable_operating_cost = pyo.Expression(
            expr=(
                (
                    self.aggregate_variable_operating_cost
                    + sum(self.aggregate_flow_costs[f] for f in self.used_flows)
                    * self.utilization_factor
                )
                if self.used_flows
                else self.aggregate_variable_operating_cost
            ),
            doc="Total variable operating cost of process per operating period",
        )

        self.total_operating_cost_constraint = pyo.Constraint(
            expr=self.total_operating_cost
            == (self.total_fixed_operating_cost + self.total_variable_operating_cost),
            doc="Total operating cost of process per operating period",
        )

        self.total_annualized_cost = pyo.Expression(
            expr=(
                self.total_capital_cost * self.capital_recovery_factor
                + self.total_operating_cost
            ),
            doc="Total annualized cost of operation",
        )

    def build_global_params(self):
        """
        Build the global parameters common to WaterTAP Costing Packages.
        The currency units should already be registered.
        """

        self.register_currency_definitions()
        self.set_base_currency_base_period()

        self.utilization_factor = pyo.Var(
            initialize=0.9,
            doc="Plant capacity utilization [fraction of uptime]",
            units=pyo.units.dimensionless,
        )

        self.electricity_cost = pyo.Var(
            initialize=0.07,
            doc="Electricity cost",
            units=pyo.units.USD_2018 / pyo.units.kWh,
        )
        self.defined_flows["electricity"] = self.electricity_cost

        self.electrical_carbon_intensity = pyo.Var(
            initialize=0.475,
            doc="Grid carbon intensity [kgCO2_eq/kWh]",
            units=pyo.units.kg / pyo.units.kWh,
        )

        self.plant_lifetime = pyo.Var(
            initialize=30, units=self.base_period, doc="Plant lifetime"
        )

        self.wacc = pyo.Var(
            # consistent with a 30 year plant_lifetime
            # and a capital_recovery_factor of 0.1
            initialize=0.09307339771758532,
            units=pyo.units.dimensionless,
            doc="Weighted Average Cost of Capital (WACC)",
        )

        self.capital_recovery_factor = pyo.Var(
            initialize=0.1,
            units=pyo.units.year**-1,
            doc="Capital annualization factor [fraction of investment cost/year]",
        )

        # used in initialize_build to check for fixed variable consistency
        self._annualization_vars = (
            self.plant_lifetime,
            self.wacc,
            self.capital_recovery_factor,
        )

        self.capital_recovery_factor_constraint = pyo.Constraint(
            expr=self.capital_recovery_factor
            == (
                (self.wacc / pyo.units.year)
                / (1 - 1 / ((1 + self.wacc) ** (self.plant_lifetime / pyo.units.year)))
            )
        )

        self.TPEC = pyo.Var(
            initialize=3.4 * (2.0 / 1.65),
            doc="Total Purchased Equipment Cost (TPEC)",
            units=pyo.units.dimensionless,
        )

        self.TIC = pyo.Var(
            initialize=2.0,
            doc="Total Installed Cost (TIC)",
            units=pyo.units.dimensionless,
        )

        self.fix_all_vars()
        self.capital_recovery_factor.unfix()

    def initialize_build(self):
        """
        Basic initialization for flowsheet level quantities
        """
        calculate_variable_from_constraint(
            self.total_capital_cost, self.total_capital_cost_constraint
        )
        calculate_variable_from_constraint(
            self.total_operating_cost, self.total_operating_cost_constraint
        )
        # handle wacc / plant_lifetime / capital_recovery_factor
        unfixed_vars = []
        for v in self._annualization_vars:
            if not v.fixed:
                unfixed_vars.append(v)
        if len(unfixed_vars) != 1:
            msg = "Exactly two of the variables "
            msg += ", ".join(v.name for v in self._annualization_vars)
            msg += " should be fixed and the other unfixed."
            raise RuntimeError(msg)
        calculate_variable_from_constraint(
            unfixed_vars[0], self.capital_recovery_factor_constraint
        )

    @staticmethod
    def add_cost_factor(blk, factor):
        """
        For a unit model costing block `blk`, adds `blk.cost_factor`,
        an expression pointing to the appropriate indirect capital
        cost multiplier, and `blk.direct_capital_cost`, which is a expression
        defined to be `blk.capital_cost / blk.cost_factor`. Valid strings for
        `factor` are "TIC" and "TPEC"; all others will result in an indirect
        capital cost factor of 1.

        Args:
            blk: an ideas.core.UnitModelCosting block
            factor: a string representing the cost factor to use
        """
        if factor == "TPEC":
            blk.cost_factor = pyo.Expression(expr=blk.costing_package.TPEC)
        elif factor == "TIC":
            blk.cost_factor = pyo.Expression(expr=blk.costing_package.TIC)
        else:
            blk.cost_factor = pyo.Expression(expr=1.0)
        blk.direct_capital_cost = pyo.Expression(
            expr=blk.capital_cost / blk.cost_factor
        )

    def _get_costing_method_for(self, unit_model):
        """
        Allow the unit model to register its default costing method,
        either through an attribute named "default_costing_method"
        or by naming the default costing method "default_costing_method"
        """
        if hasattr(unit_model, "default_costing_method"):
            return unit_model.default_costing_method
        return super()._get_costing_method_for(unit_model)

    def aggregate_costs(self):
        """
        This method aggregates costs from all the unit models and flows
        registered with this FlowsheetCostingBlock and creates aggregate
        variables for these on the FlowsheetCostingBlock that can be used for
        further process-wide costing calculations.

        The following costing variables are aggregated from all the registered
        UnitModelCostingBlocks (if they exist):

        * capital_cost,
        * direct_capital_cost,
        * fixed_operating_cost, and
        * variable_operating_cost

        Additionally, aggregate flow variables are created for all registered
        flow types along with aggregate costs associated with each of these.

        Args:
            None
        """
        super().aggregate_costs()
        c_units = self.base_currency

        @self.Expression(doc="Aggregation Expression for direct capital cost")
        def aggregate_direct_capital_cost(blk):
            e = 0
            for u in self._registered_unit_costing:
                # Allow for units that might only have a subset of cost Vars
                if hasattr(u, "direct_capital_cost"):
                    e += pyo.units.convert(u.direct_capital_cost, to_units=c_units)
                elif hasattr(u, "capital_cost"):
                    raise RuntimeError(
                        f"WaterTAP models with a capital_cost must also supply a direct_capital_cost. Found unit {u.unit_model} with `capital_cost` but no `direct_capital_cost`."
                    )

            return e

    def register_flow_type(self, flow_type, cost):
        """
        This method allows users to register new material and utility flows
        with the FlowsheetCostingBlock for use when costing flows.
        If `cost` is a constant (up to units), then this method creates a new
        `Var` on the FlowsheetCostingBlock named f`{flow_type}_cost`.
        Otherwise `cost` is a non-constant expression and this method will
        create a new `Expression` on the FlowsheetCostingBlock named
        f`{flow_type}_cost` whose value is fixed to `cost`.

        If a component named f`{flow_type}_cost` already exists on the
        FlowsheetCostingBlock, then an error is raised unless f`{flow_type}_cost`
        is `cost`. If f`{flow_type}_cost` is `cost`, no error is raised and
        the existing component f`{flow_type}_cost` is used to cost the flow.

        Args:
            flow_type: string name to represent flow type
            cost: a Pyomo expression with units representing the flow cost
        """

        flow_cost_name = flow_type + "_cost"
        current_flow_cost = self.component(flow_cost_name)
        if (current_flow_cost is None) and (not is_constant_up_to_units(cost)):
            cost_expr = pyo.Expression(expr=cost)
            self.add_component(flow_cost_name, cost_expr)
            super().register_flow_type(flow_type, cost_expr)
        else:
            # all other cases are handled in the base class
            super().register_flow_type(flow_type, cost)


@declare_process_block_class("WaterTAPCosting")
class WaterTAPCostingData(WaterTAPCostingBlockData):
    def build_global_params(self):

        # Build flowsheet level costing components
        # These are the global parameters
        self.total_investment_factor = pyo.Var(
            initialize=1.0,
            doc="Total investment factor [investment cost/equipment cost]",
            units=pyo.units.dimensionless,
        )
        self.maintenance_labor_chemical_factor = pyo.Var(
            initialize=0.03,
            doc="Maintenance-labor-chemical factor [fraction of equipment cost/year]",
            units=pyo.units.year**-1,
        )

        super().build_global_params()


@declare_process_block_class("WaterTAPCostingDetailed")
class WaterTAPCostingDetailedData(WaterTAPCostingBlockData):
    def build_global_params(self):
        """
        To minimize overhead, only create global parameters for now.
        Unit-specific parameters will be added as sub-Blocks on a case-by-case
        basis as a unit of that type is costed.
        """

        # Costing factors
        self.land_cost_percent_FCI = pyo.Var(
            units=pyo.units.dimensionless,
            doc="Land cost as % FCI",
            initialize=0.0,
        )
        self.working_capital_percent_FCI = pyo.Var(
            units=pyo.units.dimensionless,
            doc="Working capital as % FCI",
            initialize=0.0,
        )
        self.salaries_percent_FCI = pyo.Var(
            units=1 / self.base_period,
            doc="Salaries as % FCI",
            initialize=0.001 * (0.03 / 0.0149),
        )
        self.benefit_percent_of_salary = pyo.Var(
            units=pyo.units.dimensionless,
            doc="Benefits as % salaries",
            initialize=0.9,
        )
        self.maintenance_costs_percent_FCI = pyo.Var(
            units=1 / self.base_period,
            doc="Maintenance and contingency costs as % FCI",
            initialize=0.008 * (0.03 / 0.0149),
        )
        self.laboratory_fees_percent_FCI = pyo.Var(
            units=1 / self.base_period,
            doc="Laboratory fees as % FCI",
            initialize=0.003 * (0.03 / 0.0149),
        )
        self.insurance_and_taxes_percent_FCI = pyo.Var(
            units=1 / self.base_period,
            doc="Insurance and taxes as % FCI",
            initialize=0.002 * (0.03 / 0.0149),
        )

        self.total_investment_factor = pyo.Expression(
            expr=1.0 + self.working_capital_percent_FCI + self.land_cost_percent_FCI,
            doc="Total investment factor [investment cost/equipment cost]",
        )

        self.maintenance_labor_chemical_factor = pyo.Expression(
            expr=self.salaries_percent_FCI
            + self.benefit_percent_of_salary * self.salaries_percent_FCI
            + self.maintenance_costs_percent_FCI
            + self.laboratory_fees_percent_FCI
            + self.insurance_and_taxes_percent_FCI,
            doc="Maintenance-labor-chemical factor [fraction of equipment cost/year]",
        )

        super().build_global_params()

    def build_process_costs(self):
        """
        Calculating process wide costs.
        """
        # add the "basic" process costs
        super().build_process_costs()

        # now the detailed breakdowns
        # Other capital costs
        self.land_cost = pyo.Expression(
            expr=self.land_cost_percent_FCI * self.aggregate_capital_cost,
            doc="Land costs - based on aggregate capital costs",
        )
        self.working_capital = pyo.Expression(
            expr=self.working_capital_percent_FCI * self.aggregate_capital_cost,
            doc="Working capital - based on aggregate capital costs",
        )

        # Other fixed costs
        self.salary_cost = pyo.Expression(
            expr=self.salaries_percent_FCI * self.aggregate_capital_cost,
            doc="Salary costs - based on aggregate capital costs",
        )
        self.benefits_cost = pyo.Expression(
            expr=self.benefit_percent_of_salary
            * self.salaries_percent_FCI
            * self.aggregate_capital_cost,
            doc="Benefits costs - based on percentage of salary costs",
        )
        self.maintenance_cost = pyo.Expression(
            expr=self.maintenance_costs_percent_FCI * self.aggregate_capital_cost,
            doc="Maintenance costs - based on aggregate capital costs",
        )
        self.laboratory_cost = pyo.Expression(
            expr=self.laboratory_fees_percent_FCI * self.aggregate_capital_cost,
            doc="Laboratory costs - based on aggregate capital costs",
        )
        self.insurance_and_taxes_cost = pyo.Expression(
            expr=self.insurance_and_taxes_percent_FCI * self.aggregate_capital_cost,
            doc="Insurance and taxes costs - based on aggregate capital costs",
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Param,
    Constraint,
    units as pyunits,
)
from ..util import register_costing_parameter_block, make_capital_cost_var


def build_stoichiometric_reactor_cost_param_block(blk):
    blk.capital_cost_softening = Param(
        initialize=374.9,
        units=pyunits.USD_2021 / (pyunits.lb / pyunits.day),
        mutable=True,
        doc="Cost for typical mid sized softening reactor",
    )

    blk.capital_cost_acid_addition = Param(
        initialize=127.8,
        units=pyunits.USD_2021 / (pyunits.gallon / pyunits.day),
        doc="Cost for acid mixer",
    )


@register_costing_parameter_block(
    build_rule=build_stoichiometric_reactor_cost_param_block,
    parameter_block_name="stoichiometric_reactor",
)
def cost_stoichiometric_reactor(blk):
    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    if (
        blk.unit_model.has_precipitation_reaction
        and blk.unit_model.has_dissolution_reaction
    ):
        blk.capital_cost_constraint = Constraint(
            expr=blk.capital_cost
            == blk.cost_factor
            * pyunits.convert(
                blk.costing_package.stoichiometric_reactor.capital_cost_softening,
                to_units=blk.costing_package.base_currency / (pyunits.lb / pyunits.day),
            )
            * sum(
                pyunits.convert(
                    obj,
                    to_units=pyunits.lb / pyunits.day,
                )
                for reagent, obj in blk.unit_model.flow_mass_reagent.items()
            ),
        )
    elif blk.unit_model.has_dissolution_reaction:
        blk.capital_cost_constraint = Constraint(
            expr=blk.capital_cost
            == blk.cost_factor
            * pyunits.convert(
                blk.costing_package.stoichiometric_reactor.capital_cost_acid_addition,
                to_units=blk.costing_package.base_currency
                / (pyunits.gallon / pyunits.day),
            )
            * sum(
                pyunits.convert(
                    obj,
                    to_units=pyunits.gallon / pyunits.day,
                )
                for reagent, obj in blk.unit_model.flow_vol_reagent.items()
            )
        )
